# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************



from pyworkflow.utils import replaceExt, joinExt
from .mapper import Mapper
from .sqlite_db import SqliteDb

ID = 'id'
PARENT_ID = 'parent_id'
CLASSNAME = 'classname'
NAME = 'name'


class SqliteMapper(Mapper):
    """Specific Mapper implementation using Sqlite database"""
    def __init__(self, dbName, dictClasses=None):
        Mapper.__init__(self, dictClasses)
        self.__initObjDict()
        self.__initUpdateDict()
        try:
            self.db = SqliteObjectsDb(dbName)
        except Exception as ex:
            raise Exception('Error creating SqliteMapper, dbName: %s'
                            '\n error: %s' % (dbName, ex))
    
    def close(self):
        self.db.close()
        
    def commit(self):
        self.db.commit()
        
    def __getObjectValue(self, obj):
        """ Get the value of the object to be stored.
        We need to handle the special case of pointer, where we should
        store as value the object id of the pointed object.
        """
        value = obj.getObjValue()

        if obj.isPointer() and obj.hasValue():
            if value.hasObjId():  # Check the object has been stored previously
                value = value.strId()  # For pointers store the id of referenced object
            else:
                self.updatePendingPointers.append(obj)
                value = "Pending update" 
            
        return value
        
    def __insert(self, obj, namePrefix=None):
        if not hasattr(obj, '_objDoStore'):
            print("MAPPER: object '%s' doesn't seem to be an Object subclass,"
                  "       it does not have attribute '_objDoStore'. "
                  "Insert skipped." % obj)
            return
        obj._objId = self.db.insertObject(obj._objName, obj.getClassName(),
                                          self.__getObjectValue(obj),
                                          obj._objParentId,
                                          obj._objLabel, obj._objComment)
        self.updateDict[obj._objId] = obj
        sid = obj.strId()
        if namePrefix is None:
            namePrefix = sid
        else:
            namePrefix = joinExt(namePrefix, sid)
        self.insertChilds(obj, namePrefix)
        
    def insert(self, obj):
        """Insert a new object into the system, the id will be set"""
        self.__insert(obj)
        
    def insertChild(self, obj, key, attr, namePrefix=None):
        if not hasattr(attr, '_objDoStore'):
            print("MAPPER: object '%s' doesn't seem to be an Object subclass,"
                  "       it does not have attribute '_objDoStore'. \n"
                  "Insert skipped." % attr)
            return 

        if namePrefix is None:
            namePrefix = self.__getNamePrefix(obj)
        attr._objName = joinExt(namePrefix, key)
        attr._objParentId = obj._objId
        self.__insert(attr, namePrefix)
        
    def insertChilds(self, obj, namePrefix=None):
        """ Insert childs of an object, if namePrefix is None,
        the it will be deduced from obj. """
        # This is also done in insertChild, but avoid 
        # doing the same for every child element
        if namePrefix is None:
            namePrefix = self.__getNamePrefix(obj)
        for key, attr in obj.getAttributesToStore():
            self.insertChild(obj, key, attr, namePrefix)
        
    def deleteChilds(self, obj):
        namePrefix = self.__getNamePrefix(obj)
        self.db.deleteChildObjects(namePrefix)
        
    def deleteAll(self):
        """ Delete all objects stored """
        self.db.deleteAll()
                
    def delete(self, obj):
        """Delete an object and all its childs"""
        self.deleteChilds(obj)
        self.db.deleteObject(obj.getObjId())
    
    def __getNamePrefix(self, obj):
        if len(obj._objName) > 0 and '.' in obj._objName:
            return replaceExt(obj._objName, obj.strId())
        return obj.strId()
    
    def __printObj(self, obj):
        print("obj._objId", obj._objId)
        print("obj._objParentId", obj._objParentId)
        print("obj._objName", obj._objName)
        print("obj.getObjValue()", obj.getObjValue())

    def updateTo(self, obj, level=1):
        self.__initUpdateDict()
        self.__updateTo(obj, level)
        # Update pending pointers to objects
        for ptr in self.updatePendingPointers:
            self.db.updateObject(ptr._objId, ptr._objName,
                                 Mapper.getObjectPersistingClassName(ptr),
                                 self.__getObjectValue(obj), ptr._objParentId,
                                 ptr._objLabel, ptr._objComment)

        # Delete any child objects that have not been found.
        # This could be the case if some elements (such as inside List)
        # were stored in the database and were removed from the object
        self.db.deleteMissingObjectsByAncestor(self.__getNamePrefix(obj),
                                               self.updateDict.keys())

    def __updateTo(self, obj, level):
        self.db.updateObject(obj._objId, obj._objName,
                             Mapper.getObjectPersistingClassName(obj),
                             self.__getObjectValue(obj), obj._objParentId, 
                             obj._objLabel, obj._objComment)

        if obj.getObjId() in self.updateDict:
            raise Exception('Circular reference, object: %s found twice'
                            % obj.getName())
        
        self.updateDict[obj._objId] = obj

        for key, attr in obj.getAttributesToStore():
            if attr._objId is None:  # Insert new items from the previous state
                attr._objParentId = obj._objId
                namePrefix = self.__getNamePrefix(obj)
                attr._objName = joinExt(namePrefix, key)
                self.__insert(attr, namePrefix)
            else:  
                self.__updateTo(attr, level + 2)

    def updateFrom(self, obj):
        objRow = self.db.selectObjectById(obj._objId)
        self.fillObject(obj, objRow)
            
    def selectById(self, objId):
        """Build the object which id is objId"""
        if objId in self.objDict:
            obj = self.objDict[objId]
        else:
            objRow = self.db.selectObjectById(objId)
            if objRow is None:
                obj = None
            else:
                obj = self._buildObjectFromClass(objRow['classname'])
                if obj is not None:
                    self.fillObject(obj, objRow)
        return obj

    def exists(self, objId):
        return self.db.doesRowExist(objId)

    def getParent(self, obj):
        """ Retrieve the parent object of another. """
        return self.selectById(obj._objParentId)
        
    def fillObjectWithRow(self, obj, objRow):
        """ Fill the object with row data. """
        if not hasattr(obj, '_objId'):
            raise Exception("Entry '%s' (id=%s) in the database, stored as '%s'"
                            ", is being mapped to %s object. " %
                            (self._getStrValue(objRow['name']), objRow[ID],
                             objRow['classname'], type(obj)))

        obj._objId = objRow[ID]
        self.objDict[obj._objId] = obj
        obj._objName = self._getStrValue(objRow['name'])
        obj._objLabel = self._getStrValue(objRow['label'])
        obj._objComment = self._getStrValue(objRow['comment'])
        obj._objCreation = self._getStrValue(objRow['creation'])
        objValue = objRow['value']
        obj._objParentId = objRow[PARENT_ID]
        
        if obj.isPointer():
            if objValue is not None:
                objValue = self.selectById(int(objValue))
            else:
                objValue = None
        obj.set(objValue)
        
    def fillObject(self, obj, objRow, includeChildren=True):
        self.fillObjectWithRow(obj, objRow)
        namePrefix = self.__getNamePrefix(obj)

        if includeChildren:
            childs = self.db.selectObjectsByAncestor(namePrefix)

            for childRow in childs:
                childParts = childRow[NAME].split('.')
                childName = childParts[-1]
                parentId = int(childParts[-2])
                # Here we are assuming that always the parent have
                # been processed first, so it will be in the dictionary
                parentObj = self.objDict.get(parentId, None)
                if parentObj is None:  # Something went wrong
                    continue
                childObj = getattr(parentObj, childName, None)
                if childObj is None:
                    childObj = self._buildObjectFromClass(childRow[CLASSNAME])
                    # If we have any problem building the object, just ignore it
                    if childObj is None:
                        continue
                    setattr(parentObj, childName, childObj)

                self.fillObjectWithRow(childObj, childRow)

    def __buildObject(self, row):
        """ Builds and object, either based on the parent attribute, or a new
        one based on the class. """
        parentId = self._getParentIdFromRow(row)
        #  If there is no parent...
        if parentId is None:
            # It must be a Root object, use the class
            return self._buildObjectFromClass(self._getClassFromRow(row))
        else:
            # Try to get the instance from the parent
            name = self._getNameFromRow(row)
            childParts = name.split('.')
            childName = childParts[-1]

            # Get the parent, we should have it cached
            parentObj = self.objDict.get(parentId, None)
            if parentObj is None:  # Something went wrong
                print("WARNING: Parent object (id=%d) was not found, "
                      "object: %s. Ignored." % (parentId, name))
                return None

            childObj = getattr(parentObj, childName, None)

            # If parent object does not have that attribute
            if childObj is None:
                childObj = self._buildObjectFromClass(row[CLASSNAME])
                # If we have any problem building the object, just ignore it
                if childObj is None:
                    return None

                setattr(parentObj, childName, childObj)

            return childObj

    def __objFromRow(self, objRow, includeChildren=True):
        objClassName = objRow['classname']
        obj = self._buildObjectFromClass(objClassName)

        if obj is not None:
            self.fillObject(obj, objRow, includeChildren)
        
        return obj
        
    def __iterObjectsFromRows(self, objRows, objectFilter=None):
        for objRow in objRows:
            obj = self.objDict.get(objRow['id'], None) or self.__objFromRow(objRow)

            if (obj is not None and
                    objectFilter is None or objectFilter(obj)):
                yield obj
        
    def __objectsFromRows(self, objRows, iterate=False, objectFilter=None):
        """Create a set of object from a set of rows
        Params:
            objRows: rows result from a db select.
            iterate: if True, iterates over all elements, if False the whole
                list is returned
            objectFilter: function to filter some of the objects of the
                results.
        """
        if not iterate:
            return [obj for obj in self.__iterObjectsFromRows(objRows,
                                                              objectFilter)]
        else:
            return self.__iterObjectsFromRows(objRows, objectFilter)
               
    def __initObjDict(self):
        """ Clear the objDict cache """        
        self.objDict = {}
        
    def __initUpdateDict(self):
        """ Clear the updateDict cache """        
        self.updateDict = {}
        # This is used to store pointers that pointed object are not stored yet
        self.updatePendingPointers = []
         
    def selectBy(self, iterate=False, objectFilter=None, **args):
        """Select object meetings some criteria"""
        self.__initObjDict()
        objRows = self.db.selectObjectsBy(**args)
        return self.__objectsFromRows(objRows, iterate, objectFilter)
    
    def selectByClass(self, className, includeSubclasses=True, iterate=False,
                      objectFilter=None):
        self.__initObjDict()
        
        if includeSubclasses:
            from pyworkflow.utils.reflection import getSubclasses
            whereStr = "classname='%s'" % className
            base = self.dictClasses.get(className)
            subDict = getSubclasses(base, self.dictClasses)
            for k, v in subDict.items():
                if issubclass(v, base):
                    whereStr += " OR classname='%s'" % k
            objRows = self.db.selectObjectsWhere(whereStr)
            return self.__objectsFromRows(objRows, iterate, objectFilter)
        else:
            return self.selectBy(iterate=iterate, classname=className)
            
    def selectAll(self, iterate=False, objectFilter=None):
        self.__initObjDict()
        objRows = self.db.selectObjectsByParent(parent_id=None)
        return self.__objectsFromRows(objRows, iterate, objectFilter)

    def selectAllBatch(self, objectFilter=None):
        """ Select all the row at once for all the project

        Returns:
            all the protocols populated with the data from the DB

        """
        self.__initObjDict()

        # Get all the data from Objects table sorted by Name
        objAll = self.db.selectAllObjects()

        # We should have first the protocol lines
        # then each of the protocol attributes
        # at then there is the creation time

        # Dictionary to store objects
        objs = []

        # For each row
        for row in objAll:
            obj = self._getObjectFromRow(row)

            if obj is not None and objectFilter is None or objectFilter(obj):
                objs.append(obj)

        return objs

    def _getObjectFromRow(self, row):

        # Build the object without children
        obj = self.__buildObject(row)

        # If an object was created
        if obj is not None:
            # Fill object attributes with row values
            self.fillObjectWithRow(obj, row)

            # Add it to the obj cache, we might need it latter to assign
            # attributes
            self.objDict[obj._objId] = obj

            return obj

    def _getObjectFromDictionary(self, objId):
        return self.objDict[objId]

    @staticmethod
    def _getIdFromRow(row):
        return SqliteMapper._getFieldFromRow(ID, row)

    @staticmethod
    def _getParentIdFromRow(row):
        return SqliteMapper._getFieldFromRow(PARENT_ID, row)

    @staticmethod
    def _getClassFromRow(row):
        return SqliteMapper._getFieldFromRow(CLASSNAME, row)

    @staticmethod
    def _getNameFromRow(row):
        return SqliteMapper._getFieldFromRow(NAME, row)

    @staticmethod
    def _getFieldFromRow(fieldName, row):
        return row[fieldName]

    def insertRelation(self, relName, creatorObj, parentObj, childObj,
                       parentExt=None, childExt=None):
        """ This function will add a new relation between two objects.
        Params:
            relName: the name of the relation to be added.
            creatorObj: this object will be the one who register the relation.
            parentObj: this is "parent" in the relation
            childObj: this is "child" in the relation
        """
        for o in [creatorObj, parentObj, childObj]:
            if not o.hasObjId():
                raise Exception("Before adding a relation, the object should "
                                "be stored in mapper")
        self.db.insertRelation(relName, creatorObj.getObjId(),
                               parentObj.getObjId(), childObj.getObjId(),
                               parentExt, childExt)
    
    def __objectsFromIds(self, objIds):
        """Return a list of objects, given a list of id's
        """
        return [self.selectById(rowId[ID]) for rowId in objIds]
        
    def getRelationChilds(self, relName, parentObj):
        """ Return all "child" objects for a given relation.
        Params:
            relName: the name of the relation.
            parentObj: this is "parent" in the relation
        Returns: 
            a list of "child" objects.
        """
        childIds = self.db.selectRelationChilds(relName, parentObj.getObjId())
        
        return self.__objectsFromIds(childIds)  
            
    def getRelationParents(self, relName, childObj):
        """ Return all "parent" objects for a given relation.
        Params:
            relName: the name of the relation.
            childObj: this is "child" in the relation
        Returns: 
            a list of "parent" objects.
        """
        parentIds = self.db.selectRelationParents(relName, childObj.getObjId())
        
        return self.__objectsFromIds(parentIds)  

    def getRelationsByCreator(self, creatorObj):
        """ Return all relations created by creatorObj. """
        return self.db.selectRelationsByCreator(creatorObj.getObjId())
    
    def getRelationsByName(self, relationName):
        """ Return all relations stored of a given type. """
        return self.db.selectRelationsByName(relationName)

    def deleteRelations(self, creatorObj):
        """ Delete all relations created by object creatorObj """
        self.db.deleteRelationsByCreator(creatorObj.getObjId())
    
    def insertRelationData(self, relName, creatorId, parentId, childId,
                           parentExtended=None, childExtended=None):
        self.db.insertRelation(relName, creatorId, parentId, childId,
                               parentExtended, childExtended)
    
    
class SqliteObjectsDb(SqliteDb):
    """Class to handle a Sqlite database.
    It will create connection, execute queries and commands"""
    # Maintain the current version of the DB schema
    # useful for future updates and backward compatibility
    # version should be an integer number
    VERSION = 1
    
    SELECT = ("SELECT id, parent_id, name, classname, value, label, comment, "
              "datetime(creation, 'localtime') as creation FROM Objects WHERE ")
    DELETE = "DELETE FROM Objects WHERE "
    DELETE_SEQUENCE = "DELETE FROM SQLITE_SEQUENCE WHERE name='Objects'"
    
    SELECT_RELATION = ("SELECT object_%s_id AS id FROM Relations "
                       "WHERE name=? AND object_%s_id=?")
    SELECT_RELATIONS = "SELECT * FROM Relations WHERE "
    EXISTS = "SELECT EXISTS(SELECT 1 FROM Objects WHERE %s=? LIMIT 1)"
    
    def selectCmd(self, whereStr, orderByStr=' ORDER BY id'):
        return self.SELECT + whereStr + orderByStr
    
    def __init__(self, dbName, timeout=1000, pragmas=None):
        SqliteDb.__init__(self)
        self._pragmas = pragmas or {}
        self._createConnection(dbName, timeout)
        self._initialize()

    def _initialize(self):
        """ Create the required tables if needed. """
        tables = self.getTables()
        # Check if the tables have been created or not
        if not tables:
            self.__createTables()
        else:
            self.__updateTables()
        
    def __createTables(self):
        """Create required tables if don't exists"""
        # Enable foreings keys
        self.setVersion(self.VERSION)
        self._pragmas['foreing_keys'] = "ON"
        for pragma in self._pragmas.items():
            self.executeCommand("PRAGMA %s=%s" % pragma)
        # Create the Objects table
        self.executeCommand("""CREATE TABLE IF NOT EXISTS Objects
                     (id        INTEGER PRIMARY KEY AUTOINCREMENT,
                      parent_id INTEGER REFERENCES Objects(id),
                      name      TEXT,                -- object name 
                      classname TEXT,                -- object's class name
                      value     TEXT DEFAULT NULL,   -- object value, used for Scalars
                      label     TEXT DEFAULT NULL,   -- object label, text used for display
                      comment   TEXT DEFAULT NULL,   -- object comment, text used for annotations
                      creation  DATE                 -- creation date and time of the object
                      )""")
        # Create the Relations table
        self.executeCommand("""CREATE TABLE IF NOT EXISTS Relations
                     (id        INTEGER PRIMARY KEY AUTOINCREMENT,
                      parent_id INTEGER REFERENCES Objects(id), -- object that created the relation
                      name      TEXT,               -- relation name 
                      classname TEXT DEFAULT NULL,  -- relation's class name
                      value     TEXT DEFAULT NULL,  -- relation value
                      label     TEXT DEFAULT NULL,  -- relation label, text used for display
                      comment   TEXT DEFAULT NULL,  -- relation comment, text used for annotations
                      object_parent_id  INTEGER REFERENCES Objects(id) ON DELETE CASCADE,
                      object_child_id  INTEGER REFERENCES Objects(id) ON DELETE CASCADE,
                      creation  DATE,                 -- creation date and time of the object
                      object_parent_extended TEXT DEFAULT NULL, -- extended property to consider internal objects
                      object_child_extended TEXT DEFAULT NULL
                      )""")
        self.commit()
        
    def __updateTables(self):
        """ This method is intended to update the table schema
        in the case of dealing with old database version.
        """
        if self.getVersion() < self.VERSION:  # This applies for version 1
            # Add the extra column for pointer extended attribute in Relations table
            # from version 1 on, there is not needed since the table will 
            # already contains this column
            columns = [c[1] for c in self.getTableColumns('Relations')]
            if 'object_parent_extended' not in columns:
                self.executeCommand("ALTER TABLE Relations "
                                    "ADD COLUMN object_parent_extended  TEXT DEFAULT NULL")
            if 'object_child_extended' not in columns:
                self.executeCommand("ALTER TABLE Relations "
                                    "ADD COLUMN object_child_extended  TEXT DEFAULT NULL")
            self.setVersion(self.VERSION)

    def insertObject(self, name, classname, value, parent_id, label, comment):
        """Execute command to insert a new object. Return the inserted object id"""
        try:
            self.executeCommand(
                'INSERT INTO Objects (parent_id, name, classname, value, label, comment, creation)' +
                ' VALUES (?, ?, ?, ?, ?, ?, datetime(\'now\'))',
                (parent_id, name, classname, value, label, comment))
            return self.cursor.lastrowid
        except Exception as ex:
            print("insertObject: ERROR")
            print('INSERT INTO Objects (parent_id, name, classname, value, label, comment, creation)' +
                  ' VALUES (?, ?, ?, ?, ?, ?, datetime(\'now\'))')
            print((parent_id, name, classname, value, label, comment))
            raise ex
        
    def insertRelation(self, relName, parent_id, object_parent_id, object_child_id, 
                       object_parent_extended=None, object_child_extended=None, **kwargs):
        """Execute command to insert a new object. Return the inserted object id"""
        self.executeCommand("INSERT INTO Relations "
                            "(parent_id, name, object_parent_id, object_child_id, creation, "
                            "object_parent_extended, object_child_extended) "
                            " VALUES (?, ?, ?, ?, datetime('now'), ?, ?)",
                            (parent_id, relName, object_parent_id, object_child_id,
                             object_parent_extended, object_child_extended))
        return self.cursor.lastrowid
    
    def updateObject(self, objId, name, classname, value, parent_id, label, comment):
        """Update object data """
        self.executeCommand("UPDATE Objects SET parent_id=?, name=?, "
                            "classname=?, value=?, label=?, comment=? WHERE id=?",
                            (parent_id, name, classname, value, label, comment, objId))
        
    def selectObjectById(self, objId):
        """Select an object give its id"""
        self.executeCommand(self.selectCmd(ID + "=?"), (objId,))
        return self.cursor.fetchone()

    def doesRowExist(self, objId):
        """Return True if a row with a given id exists"""
        self.executeCommand(self.EXISTS % ID, (objId,))
        one = self.cursor.fetchone()
        return one[0] == 1

    def selectAllObjects(self):
        """Select all data at once"""
        self.executeCommand(self.selectCmd(ID + ">0", ' ORDER BY parent_id'))
        return self.cursor.fetchall()

    def selectObjectsByParent(self, parent_id=None, iterate=False):
        """Select object with a given parent
        if the parent_id is None, all object with parent_id NULL
        will be returned"""
        if parent_id is None:
            self.executeCommand(self.selectCmd(PARENT_ID + " is NULL"))
        else:
            self.executeCommand(self.selectCmd(PARENT_ID + "=?"), (parent_id,))
        return self._results(iterate)  
    
    def selectObjectsByAncestor(self, ancestor_namePrefix, iterate=False):
        """Select all objects in the hierarchy of ancestor_id"""
        self.executeCommand(self.selectCmd("name LIKE '%s.%%'"
                                           % ancestor_namePrefix))
        return self._results(iterate)

    def selectObjectsBy(self, iterate=False, **args):     
        """More flexible select where the constrains can be passed
        as a dictionary, the concatenation is done by an AND"""

        if len(args) == 0:
            whereStr = '1=?'
            whereTuple = (1,)
        else:
            whereList = ['%s=?' % k for k in args.keys()]
            whereStr = ' AND '.join(whereList)
            whereTuple = tuple(args.values())

        self.executeCommand(self.selectCmd(whereStr), whereTuple)
        return self._results(iterate)
    
    def selectObjectsWhere(self, whereStr, iterate=False):
        self.executeCommand(self.selectCmd(whereStr))
        return self._results(iterate)
    
    def deleteObject(self, objId):
        """Delete an existing object"""
        self.executeCommand(self.DELETE + ID + "=?", (objId,))
        
    def deleteChildObjects(self, ancestor_namePrefix):
        """ Delete from db all objects that are childs 
        of an ancestor, now them will have the same starting prefix"""
        self.executeCommand(self.DELETE + "name LIKE '%s.%%'"
                            % ancestor_namePrefix)

    def selectMissingObjectsByAncestor(self, ancestor_namePrefix,
                                       idList):
        """Select all objects in the hierarchy of ancestor_id"""
        idStr = ','.join(str(i) for i in idList)
        cmd = self.selectCmd("name LIKE '%s.%%' AND id NOT IN (%s) "
                             % (ancestor_namePrefix, idStr))
        self.executeCommand(cmd)
        return self._results(iterate=False)

    def deleteMissingObjectsByAncestor(self, ancestor_namePrefix, idList):
        """Select all objects in the hierarchy of ancestor_id"""
        idStr = ','.join(str(i) for i in idList)
        cmd = "%s name LIKE '%s.%%' AND id NOT IN (%s) " % (self.DELETE, ancestor_namePrefix, idStr)
        self.executeCommand(cmd)

    def deleteAll(self):
        """ Delete all objects from the db. """
        self.executeCommand(self.DELETE + "1")
        self.executeCommand(self.DELETE_SEQUENCE)  # restart the count of ids
        
    def selectRelationChilds(self, relName, object_parent_id):
        self.executeCommand(self.SELECT_RELATION % ('child', 'parent'), 
                            (relName, object_parent_id))
        return self._results()
        
    def selectRelationParents(self, relName, object_child_id):
        self.executeCommand(self.SELECT_RELATION % ('parent', 'child'), 
                            (relName, object_child_id))
        return self._results()
    
    def selectRelationsByCreator(self, parent_id):
        self.executeCommand(self.SELECT_RELATIONS + PARENT_ID + "=?", (parent_id,))
        return self._results()
     
    def selectRelationsByName(self, relationName):
        self.executeCommand(self.SELECT_RELATIONS + "name=?", (relationName,))
        return self._results()
       
    def deleteRelationsByCreator(self, parent_id):
        self.executeCommand("DELETE FROM Relations where parent_id=?", (parent_id,))


class SqliteFlatMapper(Mapper):
    """Specific Flat Mapper implementation using Sqlite database"""
    def __init__(self, dbName, dictClasses=None, tablePrefix='',
                 indexes=None):
        Mapper.__init__(self, dictClasses)
        self._objTemplate = None
        try:
            # We (ROB and JMRT) are playing with different
            # PRAGMAS (see https://www.sqlite.org/pragma.html)
            # for the SqliteFlatMapper instances
            # We have been playing with the following
            # pragmas: {synchronous, journal_mode, temp_store and
            #  cache_size} and inside scipion no improvement
            # has been observed. Outside scipion a careful
            #  choosing of pragmas may duplicate the speed but
            # inside scipion, I think that the overhead due
            # to the manipulation of python classes is more
            # important that the data access.

            # uncommenting these pragmas increase the speed
            # by a factor of two if NO python object is manipulated.
            # Unfortunately, any interesting operation involve
            # creation and manipulation of python objects that take
            # longer than the access time to the database
            pragmas = {
                # 0 | OFF | 1 | NORMAL | 2 | FULL | 3 | EXTRA;
                # #'synchronous': 'OFF', # ON
                # DELETE | TRUNCATE | PERSIST | MEMORY | WAL | OFF
                # #'journal_mode': 'OFF', # DELETE
                # FILE 0 | DEFAULT | 1 | FILE | 2 | MEMORY;
                # #'temp_store': 'MEMORY',
                # PRAGMA schema.cache_size = pages;
                # #'cache_size': '5000' # versus -2000
            }
            self.db = SqliteFlatDb(dbName, tablePrefix,
                                   pragmas=pragmas, indexes=indexes)
            self.doCreateTables = self.db.missingTables()
            
            if not self.doCreateTables:
                self.__loadObjDict()
        except Exception as ex:
            raise SqliteFlatMapperException('Error creating SqliteFlatMapper, '
                                            'dbName: %s, tablePrefix: %s\n error: %s' %
                                            (dbName, tablePrefix, ex))
    
    def commit(self):
        self.db.commit()
        
    def close(self):
        self.db.close()
        
    def insert(self, obj):
        if self.doCreateTables:
            self.db.createTables(obj.getObjDict(includeClass=True))
            self.doCreateTables = False
        """Insert a new object into the system, the id will be set"""
        self.db.insertObject(obj.getObjId(), obj.isEnabled(), obj.getObjLabel(), obj.getObjComment(), 
                             *obj.getObjDict().values())
        
    def enableAppend(self):
        """ This will allow to append items to existing db. 
        This is by default not allow, since most sets are not 
        modified after creation.
        """
        if not self.doCreateTables:
            obj = self.selectFirst()
            if obj is not None:
                self.db.setupCommands(obj.getObjDict(includeClass=True))
        
    def clear(self):
        self.db.clear()
        self.doCreateTables = True
    
    def deleteAll(self):
        """ Delete all objects stored """
        self.db.deleteAll()
                
    def delete(self, obj):
        """Delete an object and all its childs"""
        self.db.deleteObject(obj.getObjId())
    
    def updateTo(self, obj, level=1):
        """ Update database entry with new object values. """ 
        if self.db.INSERT_OBJECT is None:
            self.db.setupCommands(obj.getObjDict(includeClass=True))
        args = list(obj.getObjDict().values())
        args.append(obj.getObjId())
        self.db.updateObject(obj.isEnabled(), obj.getObjLabel(), obj.getObjComment(), *args)

    def exists(self, objId):
        return self.db.doesRowExist(objId)

    def selectById(self, objId):
        """Build the object which id is objId"""
        objRow = self.db.selectObjectById(objId)
        if objRow is None:
            obj = None
        else:
            obj = self.__objFromRow(objRow)
        return obj

    def __loadObjDict(self):
        """ Load object properties and classes from db. """
        # Create a template object for retrieving stored ones
        columnList = []
        rows = self.db.getClassRows()
        attrClasses = {}
        self._objBuildList = []

        for r in rows:
            label = r['label_property']

            if label == SELF:
                self._objClassName = r['class_name']
                self._objTemplate = self._buildObjectFromClass(self._objClassName)
            else:
                # Lets update the database column mapping
                self.db._columnsMapping[label] = r['column_name']
                columnList.append(label)
                attrClasses[label] = r['class_name']
                attrParts = label.split('.')
                attrJoin = ''
                o = self._objTemplate
                for a in attrParts:
                    attrJoin += a
                    attr = getattr(o, a, None)
                    if attr is None:
                        className = attrClasses[attrJoin]
                        self._objBuildList.append((className, attrJoin.split('.')))
                        attr = self._buildObjectFromClass(className)
                        setattr(o, a, attr)
                    o = attr
                    attrJoin += '.'
        basicRows = 5
        n = len(rows) + basicRows - 1
        self._objColumns = list(zip(range(basicRows, n), columnList))
         
    def __buildAndFillObj(self):
        obj = self._buildObjectFromClass(self._objClassName)
        
        for className, attrParts in self._objBuildList:
            o = obj
            for a in attrParts:
                attr = getattr(o, a, None)
                if not attr:
                    setattr(o, a, self._buildObjectFromClass(className))
                    break
                o = attr
        return obj
        
    def __objFromRow(self, objRow):
        if self._objTemplate is None:
            self.__loadObjDict()
            
        obj = self._objTemplate  # self.__buildAndFillObj()
        obj.setObjId(objRow[ID])
        obj.setObjLabel(self._getStrValue(objRow['label']))
        obj.setObjComment(self._getStrValue(objRow['comment']))
        
        try:
            obj.setEnabled(objRow['enabled'])
            obj.setObjCreation(self._getStrValue(objRow['creation']))
        except Exception:
            # THIS SHOULD NOT HAPPENS
            print("WARNING: 'creation' column not found in object: %s" % obj.getObjId())
            print("         db: %s" % self.db.getDbName())
            print("         objRow: ", dict(objRow))

        for c, attrName in self._objColumns:
            obj.setAttributeValue(attrName, objRow[c])

        return obj
        
    def __iterObjectsFromRows(self, objRows, objectFilter=None):
        for objRow in objRows:
            obj = self.__objFromRow(objRow)
            if objectFilter is None or objectFilter(obj):
                yield obj
        
    def __objectsFromRows(self, objRows, iterate=False, objectFilter=None):
        """Create a set of object from a set of rows
        Params:
            objRows: rows result from a db select.
            iterate: if True, iterates over all elements, if False the whole
                list is returned
            objectFilter: function to filter some of the objects of the results. 
        """
        if not iterate:
            return [obj.clone()
                    for obj in self.__iterObjectsFromRows(objRows, objectFilter)]
        else:
            return self.__iterObjectsFromRows(objRows, objectFilter)
         
    def selectBy(self, iterate=False, objectFilter=None, **args):
        """Select object meetings some criteria"""
        objRows = self.db.selectObjectsBy(**args)
        return self.__objectsFromRows(objRows, iterate, objectFilter)
    
    def selectAll(self, iterate=True, objectFilter=None, orderBy=ID,
                  direction='ASC', where='1', limit=None):
        # Just a sanity check for emtpy sets, that doesn't contains
        # 'Properties' table
        if not self.db.hasTable('Properties'):
            return iter([]) if iterate else []
            
        if self._objTemplate is None:
            self.__loadObjDict()
        objRows = self.db.selectAll(orderBy=orderBy,
                                    direction=direction,
                                    where=where,
                                    limit=limit)
        
        return self.__objectsFromRows(objRows, iterate, objectFilter) 

    def aggregate(self, operations, operationLabel, groupByLabels=None):
        rows = self.db.aggregate(operations, operationLabel, groupByLabels)
        results = []
        for row in rows:
            values = {}
            for label in operations:
                values[label] = row[label]
            if groupByLabels is not None:
                for label in groupByLabels:
                    values[label] = row[label]
            results.append(values)

        return results

    def count(self):
        return 0 if self.doCreateTables else self.db.count()

    def maxId(self):
        return 0 if self.doCreateTables else self.db.maxId()

    def __objectsFromIds(self, objIds):
        """Return a list of objects, given a list of id's
        """
        return [self.selectById(rowId[ID]) for rowId in objIds]
    
    def hasProperty(self, key):
        return self.db.hasProperty(key)
        
    def getProperty(self, key, defaultValue=None):
        return self.db.getProperty(key, defaultValue)
        
    def setProperty(self, key, value):
        return self.db.setProperty(key, value)
    
    def deleteProperty(self, key):
        return self.db.deleteProperty(key)
    
    def getPropertyKeys(self):
        return self.db.getPropertyKeys()
        

class SqliteFlatMapperException(Exception):
    pass


SELF = 'self'


class SqliteFlatDb(SqliteDb):
    """Class to handle a Sqlite database.
    It will create connection, execute queries and commands"""
    # Maintain the current version of the DB schema
    # useful for future updates and backward compatibility
    # version should be an integer number
    VERSION = 1
    
    CLASS_MAP = {'Integer': 'INTEGER',
                 'Float': 'REAL',
                 'Boolean': 'INTEGER'
                 }

    def __init__(self, dbName, tablePrefix='', timeout=1000,
                 pragmas=None, indexes=None):
        SqliteDb.__init__(self)
        self._pragmas = pragmas or {}
        self._indexes = indexes or []
        tablePrefix = tablePrefix.strip()
        # Avoid having _ for empty prefix
        if tablePrefix and not tablePrefix.endswith('_'):
            tablePrefix += '_'
        # NOTE (Jose Miguel, 2014/01/02
        # Reusing connections is a bit dangerous, since it have lead to
        # unexpected and hard to trace errors due to using an out-of-date
        # reused connection. That's why we are changing now the default to False
        # and only setting to True when the tablePrefix is non-empty, which is
        # the case for classes that are different tables in the same db and it
        # logical to reuse the connection.
        self._reuseConnections = bool(tablePrefix)

        self.CHECK_TABLES = ("SELECT name FROM sqlite_master WHERE type='table'"
                             " AND name='%sObjects';" % tablePrefix)
        self.SELECT = "SELECT * FROM %sObjects WHERE " % tablePrefix
        self.FROM = "FROM %sObjects" % tablePrefix
        self.DELETE = "DELETE FROM %sObjects WHERE " % tablePrefix
        self.INSERT_CLASS = ("INSERT INTO %sClasses (label_property, "
                             "column_name, class_name) VALUES (?, ?, ?)"
                             % tablePrefix)
        self.SELECT_CLASS = "SELECT * FROM %sClasses;" % tablePrefix
        self.EXISTS = "SELECT EXISTS(SELECT 1 FROM Objects WHERE %s=? LIMIT 1)"
        self.tablePrefix = tablePrefix
        self._createConnection(dbName, timeout)
        self.INSERT_OBJECT = None
        self.UPDATE_OBJECT = None
        self._columnsMapping = {}

        self.INSERT_PROPERTY = "INSERT INTO Properties (key, value) VALUES (?, ?)"
        self.DELETE_PROPERTY = "DELETE FROM Properties WHERE key=?"
        self.UPDATE_PROPERTY = "UPDATE Properties SET value=? WHERE key=?"
        self.SELECT_PROPERTY = "SELECT value FROM Properties WHERE key=?"
        self.SELECT_PROPERTY_KEYS = "SELECT key FROM Properties"

    def hasProperty(self, key):
        """ Return true if a property with this value is registered. """
        # The database not will not have the 'Properties' table when
        # there is not item inserted (ie an empty set)
        if not self.hasTable('Properties'):
            return False
        self.executeCommand(self.SELECT_PROPERTY, (key,))
        result = self.cursor.fetchone()
        return result is not None

    def getProperty(self, key, defaultValue=None):
        """ Return the value of a given property with this key.
        If not found, the defaultValue will be returned.
        """
        # The database not will not have the 'Properties' table when
        # there is not item inserted (ie an empty set)
        if not self.hasTable('Properties'):
            return defaultValue
        
        self.executeCommand(self.SELECT_PROPERTY, (key,))
        result = self.cursor.fetchone()

        if result:
            return result['value']
        else:
            return defaultValue

    def setProperty(self, key, value):
        """ Insert or update the property with a value. """
        # Just ignore the set property for empty sets
        if not self.hasTable('Properties'):
            return

        # All properties are stored as string, except for None type
        value = str(value) if value is not None else None

        if self.hasProperty(key):
            self.executeCommand(self.UPDATE_PROPERTY, (value, key))
        else:
            self.executeCommand(self.INSERT_PROPERTY, (key, value))
            
    def getPropertyKeys(self):
        """ Return all properties stored of this object. """
        self.executeCommand(self.SELECT_PROPERTY_KEYS)
        keys = [r[0] for r in self.cursor.fetchall()]
        return keys        

    def deleteProperty(self, key):
        self.executeCommand(self.DELETE_PROPERTY, (key,))

    def selectCmd(self, whereStr, orderByStr=' ORDER BY ' + ID):
        return self.SELECT + whereStr + orderByStr

    def missingTables(self):
        """ Return True is the needed Objects and Classes table are not
        created yet. """
        self.executeCommand(self.CHECK_TABLES)
        result = self.cursor.fetchone()

        return result is None

    def clear(self):
        self.executeCommand("DROP TABLE IF EXISTS Properties;")
        self.executeCommand("DROP TABLE IF EXISTS %sClasses;"
                            % self.tablePrefix)
        self.executeCommand("DROP TABLE IF EXISTS %sObjects;"
                            % self.tablePrefix)

    def createTables(self, objDict):
        """Create the Classes and Object table to store items of a Set.
        Each object will be stored in a single row.
        Each nested property of the object will be stored as a column value.
        """
        self.setVersion(self.VERSION)
        for pragma in self._pragmas.items():
            print("executing pragma", pragma)
            self.executeCommand("PRAGMA %s = %s;" % pragma)
        # Create a general Properties table to store some needed values
        self.executeCommand("""CREATE TABLE IF NOT EXISTS Properties
                     (key       TEXT UNIQUE, -- property key                 
                      value     TEXT  DEFAULT NULL -- property value
                      )""")
        # Create the Classes table to store each column name and type
        self.executeCommand("""CREATE TABLE IF NOT EXISTS %sClasses
                     (id        INTEGER PRIMARY KEY AUTOINCREMENT,
                      label_property      TEXT UNIQUE, --object label                 
                      column_name TEXT UNIQUE,
                      class_name TEXT DEFAULT NULL  -- relation's class name
                      )""" % self.tablePrefix)
        CREATE_OBJECT_TABLE = """CREATE TABLE IF NOT EXISTS %sObjects
                     (id        INTEGER PRIMARY KEY,
                      enabled   INTEGER DEFAULT 1,   -- used to selected/deselect items from a set
                      label     TEXT DEFAULT NULL,   -- object label, text used for display
                      comment   TEXT DEFAULT NULL,   -- object comment, text used for annotations
                      creation  DATE                 -- creation date and time of the object
                      """ % self.tablePrefix

        c = 0
        colMap = {}
        for k, v in objDict.items():
            colName = 'c%02d' % c
            className = v[0]
            colMap[k] = colName
            c += 1
            self.executeCommand(self.INSERT_CLASS, (k, colName, className))
            if k != SELF:
                CREATE_OBJECT_TABLE += ',%s  %s DEFAULT NULL' % (colName, self.CLASS_MAP.get(className, 'TEXT'))

        CREATE_OBJECT_TABLE += ')'
        # Create the Objects table
        self.executeCommand(CREATE_OBJECT_TABLE)

        for idx in self._indexes:
            # first check if the attribute to be indexed exists
            if idx in colMap:
                self.executeCommand("CREATE INDEX index_%s ON Objects (%s);"
                                    % (idx.replace('.', '_'), colMap[idx]))

        self.commit()
        # Prepare the INSERT and UPDATE commands
        self.setupCommands(objDict)

    def setupCommands(self, objDict):
        """ Setup the INSERT and UPDATE commands base on the object dictionary. """
        self.INSERT_OBJECT = "INSERT INTO %sObjects (id, enabled, label, comment, creation" % self.tablePrefix
        self.UPDATE_OBJECT = "UPDATE %sObjects SET enabled=?, label=?, comment=?" % self.tablePrefix
        c = 0
        for k in objDict:
            colName = 'c%02d' % c
            self._columnsMapping[k] = colName
            c += 1
            if k != SELF:
                self.INSERT_OBJECT += ',%s' % colName
                self.UPDATE_OBJECT += ', %s=?' % colName

        self.INSERT_OBJECT += ") VALUES (?,?,?,?, datetime('now')" + ',?' * (c-1) + ')'
        self.UPDATE_OBJECT += ' WHERE id=?'

    def getClassRows(self):
        """ Create a dictionary with names of the attributes
        of the colums. """
        self.executeCommand(self.SELECT_CLASS)
        return self._results(iterate=False)

    def getSelfClassName(self):
        """ Return the class name of the attribute named 'self'.
        This is the class of the items stored in a Set.
        """
        self.executeCommand(self.SELECT_CLASS)

        for classRow in self._iterResults():
            if classRow['label_property'] == SELF:
                return classRow['class_name']
        raise Exception("Row '%s' was not found in Classes table. " % SELF)

    def insertObject(self, *args):
        """Insert a new object as a row.
        *args: id, label, comment, ...
        where ... is the values of the objDict from which the tables
        where created."""
        self.executeCommand(self.INSERT_OBJECT, args)

    def updateObject(self, *args):
        """Update object data """
        self.executeCommand(self.UPDATE_OBJECT, args)

    def selectObjectById(self, objId):
        """Select an object give its id"""
        self.executeCommand(self.selectCmd(ID + "=?"), (objId,))
        return self.cursor.fetchone()

    def doesRowExist(self, objId):
        """Return True if a row with a given id exists"""
        self.executeCommand(self.EXISTS % ID, (objId,))
        one = self.cursor.fetchone()
        return one[0] == 1

    def selectAll(self, iterate=True, orderBy=ID, direction='ASC',
                  where='1', limit=None):
        # Handle the specials orderBy values of 'id' and 'RANDOM()'
        # other columns names should be mapped to table column
        # such as: _micId -> c04

        def _getRealCol(colName):
            """ Transform the column name taking into account
             special columns such as: id or RANDOM(), and
             getting the mapping translation otherwise.
            """
            if colName in ['id', 'RANDOM()', 'creation']:
                return colName
            else:
                return self._columnsMapping[colName]

        if isinstance(orderBy, str):
            orderByCol = _getRealCol(orderBy)
        elif isinstance(orderBy, list):
            orderByCol = ','.join([_getRealCol(c) for c in orderBy])
        else:
            raise Exception('Invalid type for orderBy: %s' % type(orderBy))

        # Parse the where string to replace the column name with
        # the real table column name ( for example: _micId -> c01 )
        # Right now we are assuming a simple where string in the form
        # colName=VALUE
        if '=' in where:
            whereCol = where.split('=')[0]
            whereRealCol = _getRealCol(whereCol)
            whereStr = where.replace(whereCol, whereRealCol)
        else:
            whereStr = where

        cmd = self.selectCmd(whereStr,
                             orderByStr=' ORDER BY %s %s'
                                        % (orderByCol, direction))
        # If there is a limit
        if limit:
            # if it is a tuple
            if isinstance(limit, tuple):
                limit, skipRows = limit  # Extract values from tuple
            else:
                skipRows = None
            # If we need to skip rows
            skipPart = "%s," % skipRows if skipRows else ""
            cmd += " LIMIT %s %s" % (skipPart, limit)

        self.executeCommand(cmd)
        return self._results(iterate)

    def aggregate(self, operations, operationLabel, groupByLabels=None):
        # let us count for testing
        selectStr = 'SELECT '
        separator = ' '
        # This cannot be like the following line should be expresed in terms
        # of C1, C2 etc....
        for operation in operations:
            selectStr += "%s %s(%s) AS %s" % (separator, operation,
                                              self._columnsMapping[operationLabel],
                                              operation)
            separator = ', '
        if groupByLabels is not None:
            groupByStr = 'GROUP BY '
            separator = ' '
            for groupByLabel in groupByLabels:
                groupByCol = self._columnsMapping[groupByLabel]
                selectStr += ', %(groupByCol)s as "%(groupByLabel)s"' % locals()
                groupByStr += "%s %s" % (separator, groupByCol)
                separator = ', '
        else:
            groupByStr = ' '
        sqlCommand = selectStr + "\n" + self.FROM + "\n" + groupByStr
        self.executeCommand(sqlCommand)
        return self._results(iterate=False)

    def count(self):
        """ Return the number of element in the table. """
        self.executeCommand(self.selectCmd('1').replace('*', 'COUNT(id)'))
        return self.cursor.fetchone()[0]

    def maxId(self):
        """ Return the maximum id from the Objects table. """
        self.executeCommand(self.selectCmd('1').replace('*', 'MAX(id)'))
        return self.cursor.fetchone()[0]

    # FIXME: Seems to be duplicated and a subset of selectAll
    def selectObjectsBy(self, iterate=False, **args):
        """More flexible select where the constrains can be passed
        as a dictionary, the concatenation is done by an AND"""
        whereList = ['%s=?' % k for k in args.keys()]
        whereStr = ' AND '.join(whereList)
        whereTuple = tuple(args.values())
        self.executeCommand(self.selectCmd(whereStr), whereTuple)
        return self._results(iterate)

    # FIXME: Seems to be duplicated and a subset of selectAll
    # Moreover, it does not translate between "user colums" and
    # "internal" Objects table columns
    def selectObjectsWhere(self, whereStr, iterate=False):
        self.executeCommand(self.selectCmd(whereStr))
        return self._results(iterate)

    def deleteObject(self, objId):
        """Delete an existing object"""
        self.executeCommand(self.DELETE + "id=?", (objId,))

    def deleteAll(self):
        """ Delete all objects from the db. """
        if not self.missingTables():
            self.executeCommand(self.DELETE + "1")




#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Launch main project window 
"""

import sys
import os

from pyworkflow import Config
from pyworkflow.project import Manager
from pyworkflow.gui.project import ProjectWindow
import pyworkflow.utils as pwutils


def openProject(projectName):
    """ Opens a scipion project:

    :param projectName: Name of a existing project to open,
            or "here" to create a project in the current working dir,
            or "last" to open the most recent project

    """
    manager = Manager()
    projName = os.path.basename(projectName)

    # Handle special name 'here' to create a project
    # from the current directory
    if projName == 'here':
        cwd = Config.SCIPION_CWD
        print("\nYou are trying to create a project here:",
              pwutils.cyan(cwd))

        if os.listdir(cwd):
            print(pwutils.red('\nWARNING: this folder is not empty!!!'))
        key = input("\nDo you want to create a project here? [y/N]?")

        if key.lower().strip() != 'y':
            print("\nAborting...")
            sys.exit(0)
        else:
            print("\nCreating project....")
            projName = os.path.basename(cwd)
            projDir = os.path.dirname(cwd)
            proj = manager.createProject(projName, location=projDir)

    elif projName == 'last':  # Get last project
        projects = manager.listProjects()
        if not projects:
            sys.exit("No projects yet, cannot open the last one.")
        projName = projects[0].projName

    projPath = manager.getProjectPath(projName)
    projWindow = ProjectWindow(projPath)

    projWindow.show()


if __name__ == '__main__':

    if len(sys.argv) > 1:
        openProject(sys.argv[1])
    else:
        print("usage: pw_project.py PROJECT_NAME or here or last")



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This modules handles the System management
"""

import os

import pyworkflow as pw
import pyworkflow.utils as pwutils
from .project import Project


class ProjectInfo(object):
    """Class to store some information about the project"""
    def __init__(self, projName, mTime):
        """At least it receives the Project Name and its modification time"""
        self.projName = projName
        self.mTime = mTime
        
    def getName(self):
        return self.projName
    
    def getModificationTime(self):
        return self.mTime
        
        
class Manager(object):
    """ Manage all projects of a given workspace, for a given Domain.
     (i.e, listing projects, creating, deleting or querying info)
    """
    def __init__(self, workspace=None):
        """
        Params:
            workspace: path to where the user's workspace is. A subfolder
            named 'projects' is expected. If workspace is None, the workspace
            will be taken from the global configuration.
        """
        ws = workspace or pw.Config.SCIPION_USER_DATA
        self.PROJECTS = os.path.join(ws, 'projects')

    def getProjectPath(self, projectName):
        """Return the project path given the name"""
        return os.path.join(self.PROJECTS, projectName)
        
    def listProjects(self, sortByDate=True):
        """Return a list with all existing projects
        And some other project info
        If sortByData is True, recently modified projects will be first"""
        projList = []
        pw.utils.makePath(self.PROJECTS)
        for f in os.listdir(self.PROJECTS):
            p = self.getProjectPath(f)
            if os.path.isdir(p):
                stat = os.stat(p)
                projList.append(ProjectInfo(f, stat.st_mtime))
                
        if sortByDate:
            projList.sort(key=lambda k: k.mTime, reverse=True)
        return projList
    
    def createProject(self, projectName, runsView=1, 
                      hostsConf=None, protocolsConf=None, location=None):
        """Create a new project.
        confs dict can contains customs .conf files 
        for: menus, protocols, or hosts
        """
        # If location is not None create project on it (if exists)
        if location is None:
            projectPath = self.getProjectPath(projectName)
        else:
            projectPath = os.path.join(location, projectName)

        # JMRT: Right now the project.create function change the current
        # working dir (cwd) to the project location, let's store the cwd
        # and restored after the creation
        cwd = os.getcwd()
        project = Project(pw.Config.getDomain(), projectPath)
        project.create(runsView=runsView, 
                       hostsConf=hostsConf, 
                       protocolsConf=protocolsConf)
        # If location is not the default one create a symlink on self.PROJECTS directory
        if projectPath != self.getProjectPath(projectName):
            # JMRT: Let's create the link to the absolute path, since relative
            # can be broken in systems with different mount points
            pw.utils.createAbsLink(os.path.abspath(projectPath),
                                   self.getProjectPath(projectName))

        os.chdir(cwd)  # Retore cwd before project creation

        return project

    def importProject(self, fromLocation, copyFiles=True, projectName=None, searchLocation=None):
        """ Import a project that is somewhere else in the FS
        Folder can be copied (default) or linked
        Optionally a name can be specified, otherwise name will match location folder name
        Project will always be created in the default project folder.
        """
        # If projectName is None...
        if projectName is None:
            # use same name as the import location
            projectName = os.path.basename(fromLocation)

        projectPath = self.getProjectPath(projectName)

        # If need to copyFiles
        if copyFiles:
            # Copy the whole folder
            pw.utils.copyTree(os.path.abspath(fromLocation), os.path.abspath(projectPath))

        else:
            # Link the folder
            pw.utils.createAbsLink(os.path.abspath(fromLocation), projectPath)

        project = self.loadProject(projectName)

        if searchLocation:
            project.fixLinks(searchLocation)

        return project

    def loadProject(self, projId, **kwargs):
        """ Retrieve a project object, given its id. """
        project = Project(pw.Config.getDomain(), self.getProjectPath(projId))
        project.load(**kwargs)
        return project

    def deleteProject(self, projectName):
        pw.utils.cleanPath(self.getProjectPath(projectName))

    def renameProject(self, oldName, newName):
        os.rename(self.getProjectPath(oldName), self.getProjectPath(newName))

    def hasProject(self, projectName):
        """Return True if exists a project with projectName"""
        for projInfo in self.listProjects():
            if projectName == projInfo.projName:
                return True
        return False



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Mock Domain definition that will be used by basic tests.
"""

from pyworkflow.protocol import Protocol
from pyworkflow.wizard import Wizard
from pyworkflow.viewer import Viewer
import pyworkflow.plugin as pwplugin

from .objects import MockObject


class Domain(pwplugin.Domain):
    _name = __name__
    _objectClass = MockObject
    _protocolClass = Protocol
    _viewerClass = Viewer
    _wizardClass = Wizard
    _baseClasses = globals()


class Plugin(pwplugin.Plugin):
    pass


Domain.registerPlugin(__name__)



#!/usr/bin/env python


import argparse
from pwem.viewers import EmPlotter
from pyworkflow.gui.plotter import Plotter


def main():
    parser = argparse.ArgumentParser(prog='Scipion Plot')
    parser.add_argument('--file', help='File to visualize', required=True)
    parser.add_argument('--block', help='Block to visualize')
    parser.add_argument('--type', help='Plot type')
    parser.add_argument('--columns', help='Columns to plot')
    parser.add_argument('--xcolumn', help='X Column to plot')
    parser.add_argument('--orderColumn', help='Column to order')
    parser.add_argument('--orderDir', help='Order direction(ASC, DESC)')
    parser.add_argument('--bins', help='If plot type is histogram, number of bins')
    parser.add_argument('--colors', help='Colors to plot columns')
    parser.add_argument('--styles', help='Styles to plot columns')
    parser.add_argument('--markers', help='Markers to plot columns')
    parser.add_argument('--title', help='Plot title', default='')
    parser.add_argument('--ytitle', help='Y axis title', default='')
    parser.add_argument('--xtitle', help='X axis title', default='')

    args = parser.parse_args()
    plotfile = args.file
    block = args.block if args.block else '' 
    type = args.type
    columns = args.columns
    xcolumn = args.xcolumn
    orderColumn = args.orderColumn
    orderDir = args.orderDir
    
    bins = args.bins
    colors = args.colors
    styles = args.styles
    markers = args.markers
    title = args.title
    xtitle = args.xtitle
    ytitle = args.ytitle
    
    Plotter.setBackend('TkAgg')
    plotter = EmPlotter.createFromFile(
        plotfile, block, type, columns, colors, styles, markers, xcolumn,
        ytitle, xtitle, title, bins, orderColumn, orderDir)
    plotter.show(block=True)


if __name__ == '__main__':
    main()
    
    




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

# This file is not really being used, and its tests cannot run because
# of this. So far we are not really using paramiko, and so all its
# dependencies are broken.
#
# I am commenting out all of them today. If used in the future, please
# fix. If not, please remove this file (remote.py) and test_remote.py
#
# Jordi. June 2014.

"""
This modules contains some wrappers over the ssh paramiko library
to provide handling functions on remote hosts through ssh.

Main classes are:

    RemotePath: this class use an ssh/sftp connection to mimic the
        functions in utils.path but remotely and provides basic file transfer


"""

import os
import errno
# import paramiko

from pyworkflow.utils.path import makeFilePath
from pyworkflow.utils.log import ScipionLogger

LOCAL_USER_AND_HOST = ''
SSH_PORT = 22
PAIRS_SEPARATOR = ':'

log = ScipionLogger()


def testHostConfig(host):
    """ Test the connection to a remote host give its configuration.(see HostConfig class)
    Params:
        host: configuration of the remote host, should contains hostName, userName and password.
    Returns: True if the host could be reached.
    """
    try:
        ssh = sshConnectFromHost(host)
        rpath = RemotePath(ssh)
        rpath.listdir('.')
        rpath.close()
        return True
    except Exception as ex:
        return False


def sshConnect(hostName, userName, password, port=SSH_PORT, **args):
    """ Common way to create a ssh connection.
    Params:
        hostName: Remote host name.
        userName: User name.
        password: Password.
        port: port to establish connection (usually 22)
    Returns: ssh connection handler.
    """
    raise NotImplementedError('No dependence on paramiko yet')
    # ssh = paramiko.SSHClient()
    # ssh.load_system_host_keys()
    # ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    # ssh.connect(hostName, port, userName, password, **args)
    # return ssh


def sshConnectFromHost(host):
    """ Establish a connection receiving a HostConfig class. """
    return sshConnect(host.getHostName(), host.getUserName(), host.getPassword())


class RemotePath(object):
    """ This class will server as a Wrapper to the
    paramiko sftp protocol througth a ssh connection.
    This class will implement useful methods for remote
    path handling such as: creating files, deleting folders...
    """

    @classmethod
    def fromCredentials(cls, hostName, userName, password, port=SSH_PORT, **args):
        ssh = sshConnect(hostName, userName, password, port, **args)
        rpath = RemotePath(ssh)
        return rpath

    def __init__(self, ssh):
        self.ssh = ssh
        self.sftp = ssh.open_sftp()
        # Shortcut some of the sftp methods
        self.listdir = self.sftp.listdir

    def exists(self, path):
        """ Check if a remote path exists(like os.path.exists remotely). """
        try:
            self.sftp.stat(path)
        except IOError as e:
            if e.errno == errno.ENOENT:
                return False
        else:
            return True

    def isdir(self, path):
        """ Check if a remote path is a directory (like os.path.isdir remotely). """
        try:
            from stat import S_ISDIR
            return S_ISDIR(self.sftp.stat(path).st_mode)
        except IOError:
            # Path does not exist, so by definition not a directory
            return False

    def makedirs(self, remoteFolder):
        """ Like os.makedirs remotely. """
        if len(remoteFolder) and not self.exists(remoteFolder):
            log.info('RemotePath.makedirs, path: %s' % remoteFolder)
            parent = os.path.dirname(remoteFolder)
            # if have parent and it doen't exist, create it recursively
            if len(parent) and not self.exists(parent):
                self.makedirs(parent)
            self.sftp.mkdir(remoteFolder)

    def getFile(self, remoteFile, localFile):
        """ Wrapper around sftp.get that ensures
        path exists for localFile.
        """
        makeFilePath(localFile)
        self.sftp.get(remoteFile, localFile)

    def putFile(self, localFile, remoteFile):
        """ Wrapper around sftp.put that ensures
        the remote path exists for put the file.
        """
        self.makeFilePath(remoteFile)
        log.info('RemotePath.putFile, local=%s, remote=%s' % (localFile, remoteFile))
        self.sftp.put(localFile, remoteFile)

    def makeFilePath(self, *remoteFiles):
        """ Create the remote folder path for remoteFiles. """
        self.makePath(*[os.path.dirname(r) for r in remoteFiles])

    def makePath(self, *remoteFolders):
        """ Make all path in remoteFolders list. """
        for p in remoteFolders:
            if len(p):
                self.makedirs(p)

    def close(self):
        """ Close both ssh and sftp connections. """
        self.sftp.close()
        self.ssh.close()

    def copyTree(self, localFolder, remoteFolder):
        """
        Same as shutil.copytree, but allowing
        that the dest folder also exists.
        dest is a remote folder.
        """
        if not self.exists(remoteFolder):
            self.makePath(remoteFolder)

        for path in os.listdir(localFolder):
            localPath = os.path.join(localFolder, path)
            remotePath = os.path.join(remoteFolder, path)
            # Create folders
            if os.path.isdir(localPath):
                self.copyTree(localPath, remotePath)
            elif os.path.isfile(localPath):
                self.putFile(localPath, remotePath)

    def cleanPath(self, *remotePaths):
        """
        Same as pyworkflow.utils.path.cleanPath but for remote folders or files.
        """
        for path in remotePaths:
            if self.exists(path):
                # Create folders
                if self.isdir(path):
                    files = [os.path.join(path, f) for f in self.sftp.listdir(path)]
                    self.cleanPath(*files)
                    log.info('RemotePath.cleanPath, rmdir path: %s' % path)
                    self.sftp.rmdir(path)
                else:  # os.path.isfile(path):
                    log.info('RemotePath.cleanPath, remove path: %s' % path)
                    self.sftp.remove(path)



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Data-oriented view of the project objects.
"""


import tkinter as tk
import tkinter.ttk as ttk
import tkinter.font as tkFont

import pyworkflow.utils as pwutils
import pyworkflow.protocol as pwprot
import pyworkflow.viewer as pwviewer

import pyworkflow.gui as gui
from pyworkflow.gui.tree import Tree
from pyworkflow.gui.dialog import EditObjectDialog
from pyworkflow.gui.text import TaggedText
from pyworkflow.gui import Canvas, RoundedTextBox
from pyworkflow.gui.graph import LevelTree
from pyworkflow.gui.form import getObjectLabel
from pyworkflow.constants import DATA_TAG

ACTION_EDIT = pwutils.Message.LABEL_EDIT
ACTION_COPY = pwutils.Message.LABEL_COPY
ACTION_DELETE = pwutils.Message.LABEL_DELETE
ACTION_REFRESH = pwutils.Message.LABEL_REFRESH
ACTION_STEPS = pwutils.Message.LABEL_BROWSE
ACTION_TREE = pwutils.Message.LABEL_TREE
ACTION_LIST = pwutils.Message.LABEL_LIST
ACTION_STOP = pwutils.Message.LABEL_STOP
ACTION_DEFAULT = pwutils.Message.LABEL_DEFAULT
ACTION_CONTINUE = pwutils.Message.LABEL_CONTINUE
ACTION_RESULTS = pwutils.Message.LABEL_ANALYZE

RUNS_TREE = pwutils.Icon.RUNS_TREE
RUNS_LIST = pwutils.Icon.RUNS_LIST
 
ActionIcons = {
    ACTION_EDIT: pwutils.Icon.ACTION_EDIT,
    ACTION_COPY: pwutils.Icon.ACTION_COPY,
    ACTION_DELETE:  pwutils.Icon.ACTION_DELETE,
    ACTION_REFRESH:  pwutils.Icon.ACTION_REFRESH,
    ACTION_STEPS:  pwutils.Icon.ACTION_STEPS,
    ACTION_TREE:  None,  # should be set
    ACTION_LIST:  pwutils.Icon.ACTION_LIST,
    ACTION_STOP: pwutils.Icon.ACTION_STOP,
    ACTION_CONTINUE: pwutils.Icon.ACTION_CONTINUE,
    ACTION_RESULTS: pwutils.Icon.ACTION_RESULTS,
}

STATUS_COLORS = {
    pwprot.STATUS_SAVED: '#D9F1FA',
    pwprot.STATUS_LAUNCHED: '#D9F1FA',
    pwprot.STATUS_RUNNING: '#FCCE62',
    pwprot.STATUS_FINISHED: '#D2F5CB',
    pwprot.STATUS_FAILED: '#F5CCCB',
    pwprot.STATUS_INTERACTIVE: '#F3F5CB',
    pwprot.STATUS_ABORTED: '#F5CCCB',
    # pwprot.STATUS_SAVED: '#124EB0',
}


def populateTree(tree, elements, parentId=''):
    for node in elements:
        if hasattr(node, 'count'):
            t = node.getName()
            if node.count:
                t += ' (%d)' % node.count
            node.nodeId = tree.insert(parentId, 'end', node.getName(),
                                      text=t, tags=DATA_TAG)
            populateTree(tree, node.getChilds(), node.nodeId)
            if node.count:
                tree.see(node.nodeId)
                tree.item(node.nodeId, tags=('non-empty', DATA_TAG))

    
class ProjectDataView(tk.Frame):
    def __init__(self, parent, windows, **args):
        tk.Frame.__init__(self, parent, **args)
        # Load global configuration
        self.windows = windows
        self.project = windows.project
        self.root = windows.root
        self.getImage = windows.getImage
        self.protCfg = windows.protCfg
        self.settings = windows.getSettings()
        self.showGraph = self.settings.getRunsView()
        self.style = ttk.Style()
        
        self.root.bind("<F5>", self.refreshData)
        self.__autoRefresh = None
        self.__autoRefreshCounter = 3  # start by 3 secs

        self._dataGraph = windows.project.getSourceGraph(True)

        c = self._createContent()
        
        gui.configureWeigths(self)
        c.grid(row=0, column=0, sticky='news')
        
    def _createContent(self):
        """ Create the Data View for the Project.
        It has two panes:
            Left: containing the Protocol classes tree
            Right: containing the Data list
        """
        p = tk.PanedWindow(self, orient=tk.HORIZONTAL, bg='white')
        
        # Left pane, contains Data tree
        leftFrame = tk.Frame(p, bg='white')
        bgColor = '#eaebec'
        self._createDataTree(leftFrame, bgColor)
        gui.configureWeigths(leftFrame)
        
        # Right pane
        rightFrame = tk.Frame(p, bg='white')
        rightFrame.columnconfigure(0, weight=1)
        rightFrame.rowconfigure(1, weight=1)
        # rightFrame.rowconfigure(0, minsize=label.winfo_reqheight())
        self._fillRightPane(rightFrame)

        # Add sub-windows to PanedWindows
        p.add(leftFrame, padx=5, pady=5)
        p.add(rightFrame, padx=5, pady=5)
        p.paneconfig(leftFrame, minsize=300)
        p.paneconfig(rightFrame, minsize=400)        
        
        return p
    
    def _createDataTree(self, parent, bgColor):
        """Create a tree on the left panel to store how 
        many object are from each type and the hierarchy.
        """
        defaultFont = gui.getDefaultFont()
        self.style.configure("W.Treeview",
                             background=pwutils.Color.LIGHT_GREY_COLOR,
                             borderwidth=0,
                             rowheight=defaultFont.metrics()['linespace'])
        self.dataTree = Tree(parent, show='tree', style='W.Treeview')
        self.dataTree.column('#0', minwidth=300)
        self.dataTree.tag_configure('protocol',
                                    image=self.getImage('python_file.gif'))
        self.dataTree.tag_configure('protocol_base',
                                    image=self.getImage('class_obj.gif'))
        f = tkFont.Font(family='helvetica', size='10', weight='bold')
        self.dataTree.tag_configure('non-empty', font=f)
        self.dataTree.grid(row=0, column=0, sticky='news')

        # bind click events
        self.dataTree.tag_bind(DATA_TAG, '<Double-1>', self._dataItemClick)
        self.dataTree.tag_bind(DATA_TAG, '<Return>', self._dataItemClick)
        self.dataTree.tag_bind(DATA_TAG, '<KP_Enter>', self._dataItemClick)

        # Program automatic refresh
        self.dataTree.after(3000, self._automaticRefreshData)
        
        self._updateDataTree()
        
    def _updateDataTree(self):
            
        def createClassNode(classObj):
            """ Add the object class to hierarchy and 
            any needed subclass. """
            className = classObj.__name__
            classNode = classesGraph.getNode(className)
            
            if not classNode:
                classNode = classesGraph.createNode(className)
                if className != 'EMObject' and classObj.__bases__:
                    baseClass = classObj.__bases__[0]
                    for b in classObj.__bases__:
                        if b.__name__ == 'EMObject':
                            baseClass = b
                            break
                    parent = createClassNode(baseClass)
                    parent.addChild(classNode)
                classNode.count = 0
                
            return classNode
            
        classesGraph = pwutils.Graph()
        
        self.dataTree.clear()
        for node in self._dataGraph.getNodes():
            if node.pointer:
                classNode = createClassNode(node.pointer.get().getClass())
                classNode.count += 1
        
        populateTree(self.dataTree, classesGraph.getRootNodes())

    def _fillRightPane(self, parent):
        """
        # Create the right Pane that will be composed by:
        # a Action Buttons TOOLBAR in the top
        # and another vertical Pane with:
        # Runs History (at Top)
        # Sectected run info (at Bottom)
        """
        # Create the Action Buttons TOOLBAR
        toolbar = tk.Frame(parent, bg='white')
        toolbar.grid(row=0, column=0, sticky='news')
        gui.configureWeigths(toolbar)
        # toolbar.columnconfigure(0, weight=1)
        toolbar.columnconfigure(1, weight=1)
        
        self.runsToolbar = tk.Frame(toolbar, bg='white')
        self.runsToolbar.grid(row=0, column=0, sticky='sw')
        # On the left of the toolbar will be other
        # actions that can be applied to all runs (refresh, graph view...)
        self.allToolbar = tk.Frame(toolbar, bg='white')
        self.allToolbar.grid(row=0, column=10, sticky='se')
        # self.createActionToolbar()

        # Create the Run History tree
        v = ttk.PanedWindow(parent, orient=tk.VERTICAL)
        # runsFrame = ttk.Labelframe(v, text=' History ', width=500, height=500)
        runsFrame = tk.Frame(v, bg='white')
        self._createDataGraph(runsFrame)
        gui.configureWeigths(runsFrame)
        
        # Create the Selected Run Info
        infoFrame = tk.Frame(v)
        gui.configureWeigths(infoFrame)
        self._infoText = TaggedText(infoFrame, bg='white',
                                    handlers={'sci-open': self._openProtocolFormFromId})
        self._infoText.grid(row=0, column=0, sticky='news')
        
        v.add(runsFrame, weight=3)
        v.add(infoFrame, weight=1)
        v.grid(row=1, column=0, sticky='news')

    # TODO(josemi): check that the call to RoundedTextBox is correct
    # It looks suspicious because RoundedTextBox() needs 5 arguments, not 3
    def _createNode(self, canvas, node, y):
        """ Create Data node to be painted in the graph. """
        node.selected = False
        node.box = RoundedTextBox(canvas, node, y)
        return node.box

    def _createDataItem(self, canvas, node, y):
        if node.pointer is None:
            nodeText = "Project"
        else:
            label = getObjectLabel(node.pointer, self.project.mapper)
            nodeText = (label[:25]+'...') if len(label) > 25 else label
                
        textColor = 'black'
        color = '#ADD8E6'  # Lightblue
        item = self._dataCanvas.createTextbox(nodeText, 100, y, bgColor=color,
                                              textColor=textColor)

        # Get the dataId
        if not node.isRoot():
            dataId = node.pointer.get().getObjId()

            if dataId in self.settings.dataSelection:
                item.setSelected(True)

        return item
    
    def _createDataGraph(self, parent):
        """ This will be the upper part of the right panel.
        It will contain the Data Graph with their relations.
        """
        self._dataCanvas = Canvas(parent, width=600, height=500)
        self._dataCanvas.grid(row=0, column=0, sticky='nsew')
        self._dataCanvas.onClickCallback = self._onClick
        self._dataCanvas.onDoubleClickCallback = self._onDoubleClick
        self._dataCanvas.onRightClickCallback = self._onRightClick
       
        self._updateDataGraph()
        
    def _updateDataGraph(self): 
        lt = LevelTree(self._dataGraph)
        self._dataCanvas.clear()
        lt.setCanvas(self._dataCanvas)
        lt.paint(self._createDataItem)
        self._dataCanvas.updateScrollRegion()

    def _dataItemClick(self, e=None):
        # Get the tree widget that originated the event
        # from the left pane data tree
        tree = e.widget
        dataClassName = tree.getFirst()

        if dataClassName is not None:
            self._loopData(lambda item: self._selectItemByClass(item, dataClassName))

    def _selectObject(self, pobj):
        obj = pobj.get()
        objValue = pobj.getObjValue()
        self._selected = obj
        self._infoText.setReadOnly(False)
        self._infoText.setText('*Label:* ' + getObjectLabel(pobj, self.project.mapper))
        self._infoText.addText('*Info:* ' + str(obj))
        self._infoText.addText('*Created by:*')
        self._infoText.addText(' - %s (%s)'
                               % (objValue.getObjectTag(objValue),
                                  obj.getObjCreation()))

        # Get the consumers (this will get the data!!)
        derivedData = self.project.getSourceChilds(objValue)

        if derivedData is not None and len(derivedData) > 0:
            self._infoText.addText('*Consumed by:*')

            for data in derivedData:
                # Get the protocol
                protocol = self.project.getObject(data.getObjParentId())
                self._infoText.addText(' - ' + protocol.getObjectTag(protocol))

        if obj.getObjComment():
            self._infoText.addText('*Comments:* ' + obj.getObjComment())
        self._infoText.setReadOnly(True)
        
    def _onClick(self, e=None):
        self._deselectAll()

        if e.node.pointer:
            self.toggleItemSelection(e, True)
            self._selectObject(e.node.pointer)

    def _invertSelection(self):

        self._loopData(self._invertAction)

    def _deselectAll(self):

        self._loopData(self._deselectItemAction)

    def _selectAll(self):

        self._loopData(self._selectItemAction)

    def _selectItemAction(self, item):
        self.toggleItemSelection(item, True)

    def _selectItemByClass(self, item, className):

        if not item.node.isRoot():

            data = item.node.pointer.get()
            self.toggleItemSelection(
                item, isinstance(data, self.domain.getObjects()[className]))

    def _invertAction(self, item):
        self.toggleItemSelection(item, not item.getSelected())

    def _deselectItemAction(self, item):
        self.toggleItemSelection(item, False)

    def toggleItemSelection(self, item, select):
        if item.node.isRoot():
            return
        selection = self.settings.dataSelection
        runSelection = self.settings.runSelection
        dataId = item.node.pointer.get().getObjId()
        protocolId = item.node.pointer.getObjValue().getObjId()

        if not select:
            try:
                if dataId in selection:
                    selection.remove(dataId)
                if protocolId in runSelection:
                    runSelection.remove(protocolId)
            except ValueError:
                print("id not in selection")
        else:
            selection.append(dataId)
            runSelection.append(protocolId)

        item.setSelected(select)

    def _loopData(self, action):
        results = []
        # Loop through all the items
        for key, item in self._dataCanvas.items.items():
            result = action(item)
            if result is not None:
                results.append(result)

        return results

    def _onDoubleClick(self, e=None):
        if e.node.pointer:
            self._selectObject(e.node.pointer)
            self._viewObject(e.node.pointer.get().getObjId())
            return
            # self._selectObject(e.node.pointer)
            # # Graph the first viewer available for this type of object
            # ViewerClass = findViewers(self._selected.getClassName(), DESKTOP_TKINTER)[0] #
            # viewer = ViewerClass(project=self.project)
            # viewer.visualize(self._selected)

    def _viewObject(self, objId):
        """ Call appropriate viewer for objId. """
        obj = self.project.getObject(int(objId))
        viewerClasses = self.domain.findViewers(obj.getClassName(),
                                                pwviewer.DESKTOP_TKINTER)
        if not viewerClasses:
            return  # TODO: protest nicely
        viewer = viewerClasses[0](project=self.project, parent=self.windows)
        viewer.visualize(obj)

    def _openScipionLink(self, id):
        """ So far only protocols are coming through links"""
        self._openProtocolFormFromId(id)

    def _openProtocolFormFromId(self, protId):

        prot = self.project.getObject(int(protId))
        self._openProtocolForm(prot)

    def _openProtocolForm(self, prot):
        """Open the Protocol GUI Form given a Protocol instance"""
        title = pwutils.Message.TITLE_NAME_RUN + prot.getClassName()
        w = gui.form.FormWindow(title, prot, self._executeSaveProtocol,
                                self.windows,
                                hostList=self.project.getHostNames())
        w.adjustSize()
        w.show(center=True)

    def _executeSaveProtocol(self, prot, onlySave=False):
        if onlySave:
            self.project.saveProtocol(prot)
            msg = pwutils.Message.LABEL_SAVED_FORM
            #            msg = "Protocol successfully saved."
        else:
            self.project.launchProtocol(prot)
            # Select the launched protocol to display its summary, methods..etc
            # self._selection.clear()
            # self._selection.append(prot.getObjId())
            # self._updateSelection()
            # clear lastStatus to force re-load the logs
            # self._lastStatus = None
            msg = ""

        return msg

    def _onRightClick(self, e=None):
        return [
            (pwutils.Message.LABEL_EDIT, self._editObject,
             pwutils.Icon.ACTION_EDIT),
            ('Go to protocol', self._goToProtocol,
             pwutils.Icon.ACTION_SEARCH)
        ]
    
    def _editObject(self):
        """Open the Edit GUI Form given an instance"""
        EditObjectDialog(self, pwutils.Message.TITLE_EDIT_OBJECT,
                         self._selected, self.project.mapper)

    def _goToProtocol(self):
        """Switch to protocols view selecting the correspondent protocol"""

    def refreshData(self, e=None, initRefreshCounter=True):
        """ Refresh the status of displayed data.
         Params:
            e: Tk event input
            initRefreshCounter: if True the refresh counter will be set to 3 secs
             then only case when False is from _automaticRefreshData where the
             refresh time is doubled each time to avoid refreshing too often.
        """        
        self._dataGraph = self.windows.project.getSourceGraph(True)
        self._updateDataTree()
        self._updateDataGraph()
 
        if initRefreshCounter:
            self.__autoRefreshCounter = 3  # start by 3 secs
            if self.__autoRefresh:
                self.dataTree.after_cancel(self.__autoRefresh)
                self.__autoRefresh = self.dataTree.after(
                    self.__autoRefreshCounter*1000, self._automaticRefreshData)
         
    def _automaticRefreshData(self, e=None):
        """ Schedule automatic refresh increasing the time between refreshes.
        """
        self.refreshData(initRefreshCounter=False)
        secs = self.__autoRefreshCounter
        # double the number of seconds up to 30 min
        self.__autoRefreshCounter = min(2*secs, 1800)
        self.__autoRefresh = self.dataTree.after(secs*1000,
                                                 self._automaticRefreshData)
                
                
class DataTextBox(RoundedTextBox):
    def __init__(self, canvas, text, x, y, bgColor, textColor='black'):
        RoundedTextBox.__init__(self, canvas, text, x, y, bgColor, textColor)



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Jose Gutierrez (jose.gutierrez@cnb.csic.es)
# *              Adrian Quintana (aquintana@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module defines the text used in the application.
"""


class Message:
    # Example Usage: 
    # MyMessage = Message()
    # print MyMessage.label

    # Header List
    VIEW_PROJECTS = 'Projects'
    VIEW_PROTOCOLS = 'Protocols'
    VIEW_DATA = 'Data'
    VIEW_UPLOAD = 'Upload'
    
    # Projects Template
    LABEL_PROJECTS = 'Projects'
    LABEL_CREATE_PROJECT = 'Create Project'
    LABEL_IMPORT_PROJECT = 'Import project'
    TITLE_CREATE_PROJECT = 'Enter the project name'
    TITLE_CREATE_PROJECT_NAME = 'Project Name: '
    TITLE_EDIT_OBJECT = 'Edit Object properties'
    MESSAGE_DELETE_PROJECT = 'This will *delete* the project and all its *data*. Are you sure?'
    LABEL_DELETE_PROJECT = '[Delete]'
    TITLE_DELETE_PROJECT = 'Confirm project deletion'
    LABEL_RENAME_PROJECT = '[Rename]'
    TITLE_RENAME_PROJECT = 'Confirm project renaming'
    LABEL_CREATED = 'Created: '
    LABEL_MODIFIED = 'Modified: '
    
    # Project Content Template
    LABEL_PROJECT = 'Project '

    # -- Protocol Treeview --
    LABEL_WORKFLOW = 'Workflow View: '
    LABEL_PROTTREE_NONE = 'None'

    # -- Toolbar --
    LABEL_NEW = 'New'
    LABEL_NEW_ACTION = 'New     '
    LABEL_EDIT = 'Edit'
    LABEL_RENAME = 'Rename '
    LABEL_EDIT_ACTION = 'Edit     '
    LABEL_COPY = 'Copy'
    LABEL_COPY_ACTION = 'Copy   '
    LABEL_DELETE = 'Delete'
    LABEL_DELETE_ACTION = 'Delete    '
    LABEL_STEPS = 'Steps'
    LABEL_BROWSE = 'Browse'
    LABEL_BROWSE_ACTION = 'Browse '
    LABEL_DB = 'Db'
    LABEL_STOP = 'Stop'
    LABEL_STOP_ACTION = 'Stop execution'
    LABEL_ANALYZE = 'Analyze Results'
    LABEL_TREE = 'Tree'
    LABEL_SMALLTREE = 'Small Tree'
    LABEL_LIST = 'List'
    LABEL_REFRESH = 'Refresh'
    LABEL_DEFAULT = 'Default'
    LABEL_CONTINUE = 'Continue'
    LABEL_CONTINUE_ACTION = 'Approve continue'
    LABEL_EXPORT = 'Export'
    LABEL_EXPORT_UPLOAD = 'Export & upload'
    LABEL_RESTART_WORKFLOW = 'Restart workflow'
    LABEL_CONTINUE_WORKFLOW = 'Continue workflow'
    LABEL_STOP_WORKFLOW = 'Stop from here'
    LABEL_RESET_WORKFLOW = 'Reset from here'

    # -- Tabs --
    LABEL_DATA = 'Data'
    LABEL_SUMMARY = 'Summary'
    LABEL_INPUT = 'Input'
    LABEL_OUTPUT = 'Output'
    LABEL_COMMENT = 'Comments'
    
    LABEL_OBJSUMMARY = 'Object Summary'
    LABEL_OBJINFO = 'Info'
    LABEL_OBJCREATED = 'Created'
    LABEL_OBJLABEL = 'Label'
    
    LABEL_METHODS = 'Methods'
    LABEL_BIB_BTN = 'Export references'
    LABEL_LOGS = 'Output Logs'
    LABEL_LOGS_OUTPUT = 'Output Log'
    LABEL_LOGS_ERROR = 'Error Log'
    LABEL_LOGS_SCIPION = 'Scipion Log'
    
    NO_INFO_SUMMARY = 'No summary information.'
    NO_INFO_METHODS = 'No methods information.'
    NO_INFO_LOGS = 'No logs information.'
    NO_SAVE_SETTINGS = 'Error trying to save settings.'
    
    # ------- Protocol Form messages ----------
    LABEL_CITE = 'Cite'
    LABEL_HELP = 'Help'
    TEXT_HELP = 'The file selected will be uploaded to the project folder. If the file was uploaded before, It will be replaced.'
    LABEL_RUNNAME = 'Run name'
    LABEL_EXECUTION = 'Run mode'
    LABEL_RUNMODE = 'Mode'
    LABEL_PARALLEL = 'Parallel'
    LABEL_HOST = 'Host'
    LABEL_THREADS = 'Threads'
    LABEL_MPI = 'MPI'
    LABEL_QUEUE = 'Use queue?'
    LABEL_QUEUE_FOR_JOBS = 'Use queue for Jobs?'

    LABEL_WAIT_FOR = 'Wait for'
    
    LABEL_EXPERT = 'Expert Level'
    LABEL_EXPERT_NORMAL = 'Normal'
    LABEL_EXPERT_ADVANCE = 'Advanced'
    LABEL_EXPERT_EXPERT = 'Expert'
    
    HELP_RUNMODE = """  
Normally, each protocol is composed by several atomic steps.
Each step could be computationally intensive, that's why
the *Continue* execution mode will try to continue from the
last completed step. On the other hand, the *Restart* mode
will clean the whole run directory and start from scratch.    
    """    

    HELP_MPI_THREADS = """
Define the number of processors to be used in the execution.
*MPI*: This is a number of independent processes
       that communicate through message passing
       over the network (or the same computer).
*Threads*: This refers to different execution threads 
       in the same process that can share memory. They run in
       the same computer.     
    """

    HELP_USEQUEUE = """
    Select *Yes* if you want to submit a single job per protocol to a Queue system.
    The queue commands for launch and stop jobs should be configured
    for the current host in the _hosts.conf_ file.
        """
    HELP_USEQUEUEPERJOB = """
    Select *Yes* if you want to submit the multiple jobs per protocol to a Queue system.
    The queue commands for launch and stop jobs should be configured
    for the current host in the _hosts.conf_ file.
    """

    HELP_WAIT_FOR = """
    Specify a comma separated list of protocol IDs if you want
that this protocol starts after the input protocols in the list
are finished. This function will allow you to "schedule" many
runs that will be executed after each other.
    """
    
    TITLE_NAME_RUN = ' Protocol Run: '
    TITLE_RUN = 'Run'
    TITLE_LABEL = 'Label'
    LABEL_OPT_COMMENT = 'Describe your run here...'
    TITLE_COMMENT = 'Comment'
    LABEL_RUN_MODE_RESUME = 'resume'
    LABEL_RUN_MODE_RESTART = 'restart'
    TITLE_EXEC = 'Execution'
    TITLE_BROWSE_DATA = 'Protocol data'
    LABEL_QUEUE_YES = 'Yes'
    LABEL_QUEUE_NO = 'No'
    LABEL_PARAM_YES = 'Yes'
    LABEL_PARAM_NO = 'No'
    LABEL_BUTTON_CLOSE = 'Close'
    LABEL_BUTTON_SAVE = 'Save'
    LABEL_BUTTON_EXEC = 'Execute'
    LABEL_BUTTON_VIS = 'Visualize'
    LABEL_BUTTON_WIZ = 'Wizard'
    LABEL_BUTTON_HELP = 'Help'
    LABEL_BUTTON_RETURN = 'Save'
    # VARS
    VAR_EXEC_HOST = 'hostName'
    VAR_EXPERT = 'expertLevel'
    VAR_MPI = 'numberOfMpi'
    VAR_QUEUE = '_useQueue'
    VAR_RUN_NAME = 'runName'
    VAR_COMMENT = 'comment'
    VAR_RUN_MODE = 'runMode'
    VAR_THREADS = 'numberOfThreads'
    
    LABEL_PATTERN = 'Pattern'
    TEXT_PATTERN = """\
Pattern (that can include wildcards) of the files to import.
For example:
  *data/particles/***.spi*
  *~/Micrographs/mic/***.mrc*"""
    ERROR_PATTERN_EMPTY = 'The *pattern* cannot be empty.'
    ERROR_PATTERN_FILES = 'There are no files matching the *pattern*'
    LABEL_CHECKSTACK = 'Check stack files?'
    LABEL_COPYFILES = 'Copy files?'
    LABEL_VOLTAGE = 'Microscope voltage (kV)'
    TEXT_VOLTAGE = "Microscope voltage"
    LABEL_SPH_ABERRATION = 'Spherical aberration (mm)'
    TEXT_SPH_ABERRATION = """\
Optical effect due to the increased refraction of light rays when they
strike the lens near its edge, in comparison with those that strike near
the center."""
    LABEL_AMPLITUDE = 'Amplitude Contrast'
    TEXT_AMPLITUDE = """\
Produced by the loss of amplitude (i.e. electrons) from the beam.

For a weak phase and weak amplitude object, the amplitude contrast ratio Qo
is automatically computed. It should be a positive number, typically between
0.05 and 0.3."""
    LABEL_PATTERNU = 'Pattern untilted'
    LABEL_PATTERNT = 'Pattern tilted'

    LABEL_SAMP_MODE = 'Sampling rate mode'
    TEXT_SAMP_MODE = """\
You can specify the sampling rate (pixel size) directly from the image
(A/pixel, Ts) or by specifying the magnification rate (M) and the scanner
pixel size (microns/pixel, Tm).

They are related by  Ts = Tm / M"""
    LABEL_SAMP_MODE_1 = 'From image'
    LABEL_SAMP_MODE_2 = 'From scanner'
    LABEL_SAMP_RATE = 'Pixel size ("sampling rate") (Å/px)'
    TEXT_SAMP_RATE = "Pixel size"
    LABEL_MAGNI_RATE = 'Magnification rate'
    TEXT_MAGNI_RATE = """\
Electron optical magnification (M). It can be used to compute the Image Pixel
Size ("Sampling Rate") (Ts) using the Scanner Pixel Size (Tm), Ts = Tm / M.

It is used by a few programs like Ctffind or Frealign."""
    LABEL_SCANNED = 'Scanned pixel size (microns/px)'

    ERROR_IMPORT_VOL = 'importVolumes:There is not filePaths matching pattern'
    
    LABEL_CTF_ESTI = 'CTF Estimation'
    LABEL_INPUT_MIC = 'Input Micrographs'
    LABEL_INPUT_PART = 'Input Particles'
    LABEL_INPUT_VOLS = 'Input Volumes'
    LABEL_INPUT_MOVS = 'Input Movies'
    LABEL_ALIG_PART = 'Write aligned particles?'
    TEXT_ALIG_PART = 'If set to *Yes*, the alignment will be applied to \n'+'input particles and a new aligned set will be created.'

    TEXT_NO_INPUT_MIC = 'No *Input Micrographs* selected.'
    TEXT_NO_CTF_READY = 'CTF of *Input Micrographs* not ready yet.'
    TEXT_NO_OUTPUT_CO = 'Output coordinates not ready yet.'
    ERROR_NO_EST_CTF = '_estimateCTF should be implemented'
    TEXT_NO_OUTPUT_FILES = 'No output file produced'
    
    TITLE_LAUNCHED = 'Success'
    LABEL_LAUNCHED = 'The protocol was launched successfully.'
    LABEL_FOUND_ERROR = 'Errors found'
    TITLE_SAVED_FORM = 'Success'
    LABEL_SAVED_FORM = 'The protocol was saved successfully.'
    TITLE_DELETE_FORM = 'Confirm DELETE'
    TITLE_RESTART_FORM = 'Confirm RESTART'
    LABEL_DELETE_FORM = """
You are going to *DELETE* the run(s): 
  - %s
*ALL DATA* related will be permanently removed.

Do you really want to continue?'
"""
    LABEL_RESTART_FORM = """
You are going to *RESTART* the run: 
  %s

Do you really want to restart?'
"""
    LABEL_RESTART_PROTOCOL = """
    Do you want to *RESTART* ONLY the run: 
      %s 
    *ALL DATA* related with the rest of protocols that depends on this, 
    will be permanently removed.  
    """
    TITLE_STOP_FORM = 'Confirm STOP'
    LABEL_STOP_FORM = 'Do you really want to *STOP* this run?'
    
    NO_VIEWER_FOUND = 'There is not viewer for protocol:' 
    
    TITLE_SAVE_OUTPUT = 'Save protocol output'
    LABEL_SAVE_OUTPUT = 'Do you wish to save protocol output?'

    TITLE_RESTART_WORKFLOW_FORM = 'Confirm RESTART'
    TITLE_STOP_WORKFLOW_FORM = 'Confirm STOP'
    TITLE_RESET_WORKFLOW_FORM = 'Confirm RESET'
    TITLE_RESTART_WORKFLOW = 'Do you really want to *RESTART* this Workflow? All previous results will be deleted'
    TITLE_STOP_WORKFLOW = 'Do you really want to *STOP* this Workflow?'
    TITLE_RESET_WORKFLOW = 'Do you really want to *RESET* this Workflow?'
    TITLE_LAUNCHED_WORKFLOW_FAILED_FORM = 'Error while launching the workflow'
    TITLE_STOPPED_WORKFLOW_FAILED = 'Error while stopping the workflow'
    TITLE_RESETED_WORKFLOW_FAILED = 'Error while resetting the workflow'
    TITLE_LAUNCHED_WORKFLOW_FAILED = 'The workflow can not be relaunch from this protocol.\n'
    
    # SHOWJ_WEB
    SHOWJ_TITLE = 'Showj'
    
    LABEL_RESLICE = 'Reslice'
    RESLICE_Z_NEGATIVE = 'Z Negative (Front)'
    RESLICE_Y_NEGATIVE = 'Y Negative (Top)'
    RESLICE_Y_POSITIVE = 'Y Positive (Bottom)'
    RESLICE_X_NEGATIVE = 'X Negative (Left)'
    RESLICE_X_POSITIVE = 'X Positive (Right)'
    
    LABEL_COLS = 'Cols'
    LABEL_ROWS = 'Rows'
    
    LABEL_MIRRORY = 'Invert Y Axis'
    LABEL_APPLY_TRANSFORM = 'Apply Transform Matrix'
    LABEL_ONLY_SHIFTS = 'Only Shifts'
    LABEL_WRAP = 'Wrap'
    
    LABEL_BLOCK_SELECTION = 'Select Block'
    LABEL_LABEL_SELECTION = 'Select Label'
    LABEL_VOLUME_SELECTION = 'Select Volume'
    
    LABEL_ENABLE = 'Enable'
    LABEL_DISABLE = 'Disable'
    LABEL_SELECT_ALL = 'Select all'
    LABEL_SELECT_FROM = 'Select from here'
    LABEL_SELECT_TO = 'Select to here'
    
    LABEL_DISPLAY_TABLE_CONFIG = 'Display Table Configuration'
    
    LABEL_LABEL = 'Label'
    LABEL_VISIBLE = 'Visible'
    LABEL_RENDER = 'Render'
    
    LABEL_BUTTON_OK = 'Ok'
    LABEL_BUTTON_CANCEL = 'Cancel'
    
    LABEL_THRESHOLD = 'Threshold:'
    
    ERROR_DIMENSIONS = 'Incorrect table width or height: '
    ERROR_WEBGL = 'Your web browser does not support or is not configured for WebGL. See [[http://get.webgl.org/][WebGL Support]] for more information.'
    
    TOOLTIP_SEARCH = 'Search a given world in the text. '
    TOOLTIP_SEARCH_NEXT = 'Move to the next highlighted item. Also, press <Down> or <F3>'
    TOOLTIP_SEARCH_PREVIOUS = 'Move to the previous highlighted item. Also, press <Up> or <Shift-F3>'
    TOOLTIP_REFRESH = 'Reload the content of the files in the viewer. '
    TOOLTIP_EXTERNAL = 'Open the viewer in an external window. '

    TITLE_PICK_GAUSS = 'Automatic gaussian picking'
    LABEL_PICK_GAUSS = 'Do you wish to perform an automatic gaussian picking for the remaining micrographs?'

    TITLE_INSPECTOR = 'Objects inspector'
    LABEL_INSPECTOR = 'Objects inspector will inspect the whole project. ' \
                      'Thus, it can take a while depending on the size of the project.\n' \
                      'Do you want to continue?'
    EXECUTE_PLUGINS_MANAGER_OPERATION = 'Execute all selected operations'
    CANCEL_SELECTED_OPERATION = 'Cancel a selected operation'


# PLUGIN/BINARY STATES
class PluginStates:
    PLUGIN = 'plugin'
    BINARY = 'binary'
    UNCHECKED = 'unchecked'
    CHECKED = 'checked'
    INSTALL = 'install'
    UNINSTALL = 'uninstall'
    TO_INSTALL = 'to_install'
    INSTALLED = 'installed'
    PRECESSING = 'processing'
    FAILURE = 'failure'
    AVAILABLE_RELEASE = 'available_release'
    TO_UPDATE = 'to_update'
    SUCCESS = 'success'
    ERRORS = 'errors'


# To get font awesome icons into png use: http://fa2png.io/
class Icon:
    # Project Content Template
    RUNS_TREE = 'fa-sitemap.gif'
    RUNS_LIST = 'fa-bars.gif'
    ACTION_NEW = 'fa-plus-circle.gif'
    ACTION_EDIT = 'fa-pencil.gif'
    ACTION_SELECT_TO = 'fa-arrow-up.gif'
    ACTION_COPY = 'fa-files-o.gif'
    ACTION_DELETE = 'fa-trash-o.gif'
    ACTION_REFRESH = 'fa-refresh.gif'
    ACTION_RENAME = 'rename.gif'
    # TODO: change action_steps icon - fa-codefork?
    ACTION_STEPS = 'fa-list-ul.gif'
    ACTION_BROWSE = 'fa-folder-open.gif'
    ACTION_DB = 'fa-database.gif'
    ACTION_TREE = None
    ACTION_LIST = 'fa-bars.gif'
    ACTION_STOP = 'fa-stop.gif'
    ACTION_CONTINUE = 'fa-play-circle-o.gif'
    ACTION_STOP_WORKFLOW = 'fa-stop-workflow.gif'
    ACTION_RESULTS = 'fa-eye.gif'
    ACTION_CLOSE = 'fa-times.gif'
    ACTION_SAVE = 'fa-save.gif'
    ACTION_VISUALIZE = 'fa-eye.gif'
    ACTION_WIZ = 'fa-magic.gif'
    ACTION_HELP = 'fa-question-circle.gif'
    ACTION_REFERENCES = 'fa-external-link.gif'
    ACTION_EXPORT = 'fa-external-link.gif'
    ACTION_EXPORT_UPLOAD = 'fa-upload.gif'

    ACTION_SEARCH = 'fa-search.gif'
    ACTION_EXECUTE = 'fa-cogs.gif'
    ACTION_IN = 'fa-sign-in.gif'
    ACTION_OUT = 'fa-sign-out.gif'
    ACTION_FIND_NEXT = 'fa-next.gif'
    ACTION_FIND_PREVIOUS = 'fa-previous.gif'
    # Host template
    BUTTON_SELECT = 'fa-check.gif'
    BUTTON_CLOSE = 'fa-times.gif'
    BUTTON_CANCEL = 'fa-ban.gif'
    BUTTON_SAVE = ACTION_SAVE
    BUTTON_PC = 'fa-laptop.gif'
    
    ARROW_UP = 'fa-arrow-up.gif'
    ARROW_LEFT = 'fa-arrow-left.gif'
    BRUSH = 'fa-paint-brush.gif'
    TAGS = 'fa-tags.gif'
    HOME = 'fa-home.gif'
    LIGHTBULB = 'fa-lightbulb-o.gif'
    PLUS_CIRCLE = 'fa-plus-circle.gif'
    LINK = 'fa-iconmoon-link.gif'
    ROOT = 'root.gif'
    ROCKET = 'fa-rocket.gif'

    SCIPION_ICON = 'scipion_icon.gif'
    SCIPION_ICON_PROJ = SCIPION_ICON  # 'scipion_icon_proj.gif'
    SCIPION_ICON_PROJS = SCIPION_ICON  # 'scipion_icon_projs.gif'
    SCIPION_ICON_PROT = SCIPION_ICON  # 'scipion_icon_prot.gif'

    # PLUGIN MANAGER ICONS
    CHECKED = 'fa-checked.gif'
    UNCHECKED = 'fa-unchecked.gif'
    INSTALL = 'fa-install.gif'
    UNINSTALL = 'fa-uninstall.gif'
    TO_INSTALL = 'fa-to_install.gif'
    INSTALLED = 'fa-installed.gif'
    PROCESSING = 'fa-processing.gif'
    FAILURE = 'fa-failure.gif'
    DELETE_OPERATION = 'fa-delete-operation.gif'
    TO_UPDATE = 'fa-update.gif'
    ACTION_UNDO = 'fa-undo.gif'


class Color:
    RED_COLOR = 'Firebrick'  # Red color for background label  = #B22222
    LIGHT_RED_COLOR = '#F3CBCB'  # Very light red for row selection
    LIGHT_BLUE_COLOR = '#EAEBFF'  # Very light blue for even rows
    LIGHT_GREY_COLOR = '#EAEBEC'  # Light grey for background color in form, protocol, table header and west container
    LIGHT_GREY_COLOR_2 = '#F2F2F2'  # Very light grey for odd rows, input background, etc
    DARK_GREY_COLOR = '#6E6E6E'  # Very dark grey for project title, tubes, etc
    
    STATUS_SAVED = '#D9F1FA', 
    STATUS_LAUNCHED = '#D9F1FA', 
    STATUS_RUNNING = '#FCCE62', 
    STATUS_FINISHED = '#D2F5CB', 
    STATUS_FAILED = '#F5CCCB', 
    STATUS_INTERACTIVE = '#F3F5CB',
    STATUS_ABORTED = '#F5CCCB',


class colorText:
    """printing in colors, bold, etc,
       example: print colorText.BOLD + 'Hello World !' + color.END
    """
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


class KEYSYM:
    """ Keysym values for evaluating key pressed within events
    as reported at http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/key-names.html
    """
    DELETE = 'Delete'
    RETURN = 'Return'



# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es) [1]
# *              J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [2]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# * [2] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow as pw
import pyworkflow.object as pwobj
import pyworkflow.tests as pwtests
import pyworkflow.mapper as pwmapper
import pyworkflow.protocol as pwprot
from pyworkflowtests.protocols import ProtOutputTest
from pyworkflowtests import Domain, MockObject

# Protocol to output of basic scipion objects
class TestProtocolOutputs(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls, writeLocalConfig=True)
        pwtests.setupTestOutput(cls)

        # Set the application domain
        pw.Config.setDomain("pyworkflowtests")

    def test_basicObjectOutput(self):
        """Test the list with several Complex"""
        fn = self.getOutputPath("protocol.sqlite")

        # Discover objects and protocols
        mapperDict = Domain.getMapperDict()

        mapper = pwmapper.SqliteMapper(fn, mapperDict)
        prot = ProtOutputTest(mapper=mapper, n=2,
                              workingDir=self.getOutputPath(''))

        # Add and old style o, not in the outputs dictionary
        prot.output1 = MockObject()

        self.assertFalse(prot._useOutputList.get(),
                         "useOutputList wrongly initialized")

        outputs = [o for o in prot.iterOutputAttributes()]
        self.assertTrue(1, len(outputs))

        prot._stepsExecutor = pwprot.StepExecutor(hostConfig=None)
        prot.run()

        self.assertEqual(prot._steps[0].getStatus(),
                         pwprot.STATUS_FINISHED)

        # Check there is an o
        self.assertOutput(prot)

        outputs = [o for o in prot.iterOutputAttributes()]

        # We are intentionally ignoring a protocol with o (EMObject)
        # That has been continued, We do not find a real case now.
        self.assertEqual(1, len(outputs),
                         msg="Integer o not registered properly.")

        outputs = [o for o in prot.iterOutputAttributes(pwobj.Integer)]

        # Test passing a filter
        self.assertEqual(1, len(outputs),
                         msg="Integer not matched when filtering outputs.")

        # Test with non existing class
        class NotRealClass:
            pass

        outputs = [o for o in prot.iterOutputAttributes(NotRealClass)]

        # Test passing a class
        self.assertEqual(0, len(outputs),
                         msg="Filter by class in iterOutputAttributes does "
                             "not work.")

        self.assertTrue(prot._useOutputList.get(),
                        "useOutputList not activated")

    def test_basicObjectInProject(self):

        prot = self.newProtocol(ProtOutputTest,
                                objLabel='to generate basic input')
        print("working dir: %s" % prot.getWorkingDir())
        # Define a negative output for later tests
        prot._defineOutputs(negative=pwobj.Integer(-20))
        self.launchProtocol(prot)

        # Default value is 10 so output is 20
        self.assertOutput(prot)

        # Second protocol to test linking
        prot2 = self.newProtocol(ProtOutputTest,
                                 objLabel='to read basic input')

        # Set the pointer for the integer
        prot2.iBoxSize.setPointer(pwobj.Pointer(prot, extended="oBoxSize"))
        self.launchProtocol(prot2)
        self.assertOutput(prot2, value=40)

        # Test validation: only positive numbers are allowed
        prot3 = self.newProtocol(ProtOutputTest,
                                 objLabel='invalid input',
                                 iBoxSize=-10)
        # We expect this to fail
        with self.assertRaises(Exception):
            self.launchProtocol(prot3)
        # Test validation: pointer value is validated
        prot4 = self.newProtocol(ProtOutputTest,
                                 objLabel='invalid pointer input')
        # Now use negative pointer output
        prot4.iBoxSize.setPointer(pwobj.Pointer(prot, extended="negative"))

        # We expect this to fail
        with self.assertRaises(Exception):
            self.launchProtocol(prot4)

    def assertOutput(self, prot, value=20):
        # Check there is an output
        self.assertTrue(hasattr(prot, 'oBoxSize'),
                        msg="Protocol output boxSize (OInteger) not registered"
                            " as attribute.")

        self.assertEqual(value, prot.oBoxSize.get(),
                         "oBoxSize value is wrong: %s , expected %s" %
                         (prot.oBoxSize, value))



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************



import os
import tkinter as tk
import tkinter.font as tkFont

import pyworkflow as pw
from pyworkflow.project import Project
from pyworkflow.utils.utils import prettyDate, prettyTime
from pyworkflow.utils.path import getHomePath

from pyworkflow.project import Manager
import pyworkflow.gui as pwgui
from pyworkflow.gui.text import TaggedText
from pyworkflow.gui.dialog import askString, askYesNo, showError

from pyworkflow.gui import Message, Window, cfgEntryBgColor
from pyworkflow.gui.browser import FileBrowserWindow
from pyworkflow.gui.widgets import IconButton, HotButton, Button
from pyworkflow.utils.properties import Icon

            
class ProjectsView(tk.Frame):
    def __init__(self, parent, windows, **args):
        tk.Frame.__init__(self, parent, bg='white', **args)
        self.windows = windows
        self.manager = windows.manager
        self.root = windows.root

        # tkFont.Font(size=12, family='verdana', weight='bold')
        bigSize = pwgui.cfgFontSize + 2
        smallSize = pwgui.cfgFontSize - 2
        fontName = pwgui.cfgFontName

        self.projNameFont = tkFont.Font(size=bigSize, family=fontName,
                                        weight='bold')
        self.projDateFont = tkFont.Font(size=smallSize, family=fontName)
        self.projDelFont = tkFont.Font(size=smallSize, family=fontName,
                                       weight='bold')
        self.manager = Manager()

        self.filter = tk.StringVar()
        self.filterBox = None
        self.addActionsFrame()

        self.columnconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        text = TaggedText(self, width=40, height=15, bd=0, bg='white')
        text.grid(row=1, columnspan=2, column=0, sticky='news')

        self.createProjectList(text)
        text.setReadOnly(True)
        self.text = text
        self.filterBox.focus_set()

    def addActionsFrame(self):
        """ Add the "toolbar" for actions like create project, import
         project or filter"""
        # Add the create project button
        bg = "white"
        btnFrame = tk.Frame(self, bg=bg)
        btn = HotButton(btnFrame, text=Message.LABEL_CREATE_PROJECT,
                        font=self.projNameFont,
                        command=self._onCreateProject)
        btn.grid(row=0, column=0, sticky='nw', padx=10, pady=10)
        # Add the Import project button
        btn = Button(btnFrame, text=Message.LABEL_IMPORT_PROJECT,
                     font=self.projNameFont,
                     command=self._onImportProject)
        btn.grid(row=0, column=1, sticky='nw', padx=10, pady=10)
        btnFrame.grid(row=0, column=0, sticky='nw')

        # Add a filter box
        # Add the Import project button
        btn = tk.Label(btnFrame, bg=bg, text="Filter:", font=self.projNameFont)
        btn.grid(row=0, column=2, sticky='nse', padx=10, pady=10)
        self.filterBox = tk.Entry(btnFrame, font=self.projNameFont, textvariable=self.filter)
        self.filterBox.grid(row=0, column=3, sticky='ne', padx=10, pady=12)
        self.filterBox.bind('<Return>', self._onFilter)
        self.filterBox.bind('<KP_Enter>', self._onFilter)

    def createProjectList(self, text):
        """Load the list of projects"""
        r = 0
        text.setReadOnly(False)
        text.clear()
        parent = tk.Frame(text, bg='white')
        parent.columnconfigure(0, weight=1)
        colors = ['white', '#EAEBFF']
        for i, p in enumerate(self.manager.listProjects()):
            try:
                project = self.manager.loadProject(p.getName(), chdir=False, loadAllConfig=False)
                project.closeMapper()
                # Add creation time to project info
                p.cTime = project.getCreationTime()
                # Add if it's a link
                p.isLink = project.isLink()

                # Consider the filter
                if not self._doesProjectMatchFilter(p):
                    continue

                # If it's a link, get the linked folder
                if p.isLink:
                    p.linkedFolder = os.path.realpath(project.path)
                frame = self.createProjectLabel(parent, p, color=colors[i % 2])
                frame.grid(row=r, column=0, padx=10, pady=5, sticky='new')
                r += 1
            except Exception as ex:
                print("ERROR loading project: %s" % p.getName())
                print(ex)
        text.window_create(tk.INSERT, window=parent)
        text.bindWidget(parent)
        text.setReadOnly(True)

    def createProjectLabel(self, parent, projInfo, color):
        frame = tk.Frame(parent, bg=color)
        # ROW1
        # Project name
        label = tk.Label(frame, text=projInfo.projName, anchor='nw', bg=color,
                         justify=tk.LEFT, font=self.projNameFont, cursor='hand1', width=50)
        label.grid(row=0, column=0,  padx=2, pady=2, sticky='nw')
        label.bind('<Button-1>', lambda e: self.openProject(projInfo.projName))

        # ROW2
        # Timestamp line
        dateMsg = '%s%s    %s%s' % (Message.LABEL_MODIFIED, prettyDate(projInfo.mTime),
                                    Message.LABEL_CREATED, prettyTime(projInfo.cTime, time=False))
        dateLabel = tk.Label(frame, text=dateMsg, font=self.projDateFont, bg=color)
        dateLabel.grid(row=1, column=0, sticky='nw')
        # Delete action
        delLabel = tk.Label(frame, text=Message.LABEL_DELETE_PROJECT, font=self.projDelFont, bg=color, cursor='hand1')
        delLabel.grid(row=1, column=1, padx=10)
        delLabel.bind('<Button-1>', lambda e: self.deleteProject(projInfo.projName))
        # Rename action
        mvLabel = tk.Label(frame, text=Message.LABEL_RENAME_PROJECT, font=self.projDelFont, bg=color, cursor='hand1')
        mvLabel.grid(row=1, column=2)
        mvLabel.bind('<Button-1>', lambda e: self.renameProject(projInfo.projName))

        # ROW3
        if projInfo.isLink:
            linkMsg = 'link --> ' + projInfo.linkedFolder
            lblLink = tk.Label(frame, text=linkMsg, font=self.projDateFont, bg=color, fg='grey', justify=tk.LEFT)
            lblLink.grid(row=2, column=0, columnspan=3, sticky='w')

        return frame

    def createNewProject(self, projName, projLocation):
        self.manager.createProject(projName, location=projLocation)
        self.createProjectList(self.text)
        self.openProject(projName)

    def _onCreateProject(self, e=None):
        projWindow = ProjectCreateWindow("Create project", self)
        projWindow.show()

    def _onImportProject(self, e=None):
        importProjWindow = ProjectImportWindow("Import project", self)
        importProjWindow.show()

    def _onFilter(self, e=None):
        self.createProjectList(self.text)

    def _setFocusToList(self, e=None):
        self.text.focus_set()

    def _doesProjectMatchFilter(self, project):
        """ Returns true if the project matches the filter"""
        # Lets' use the name anc creation date for now:
        searchString = "~".join([project.getName().lower(),
                                 prettyDate(project.mTime),
                                 prettyTime(project.cTime, time=False)])

        return self.filter.get().lower() in searchString


    def importProject(self, projLocation, copyFiles, projName, searchLocation):

        self.manager.importProject(projLocation, copyFiles, projName, searchLocation)
        self.createProjectList(self.text)
        self.openProject(projName)

    def openProject(self, projName):
        from subprocess import Popen
        script = pw.join(pw.APPS, 'pw_project.py')
        Popen([pw.PYTHON, script, projName])

    def deleteProject(self, projName):
        if askYesNo(Message.TITLE_DELETE_PROJECT,
                    "Project *%s*. " % projName + Message.MESSAGE_DELETE_PROJECT, self.root):
            self.manager.deleteProject(projName)
            self.createProjectList(self.text)

    def renameProject(self, projName):
        newName = askString("Rename project %s" % projName, "Enter new name:", self.root)
        if not newName or newName == projName:
            return
        if self.manager.hasProject(newName):
            showError("Rename cancelled",
                      "Project name already exists: %s" % newName, self.root)
            return
        self.manager.renameProject(projName, newName)
        self.createProjectList(self.text)


class ProjectCreateWindow(Window):
    """ Windows to create a project. """
    def __init__(self, title, parent=None, weight=True, minsize=(400, 110),
                 icon=Icon.SCIPION_ICON, **args):
        """
         We assume the parent should be of ProjectsView
        """
        Window.__init__(self, title, parent.windows, weight=weight,
                        icon=icon, minsize=minsize, enableQueue=True)
        self.root['background'] = 'white'

        self.parent = parent
        self.projectsPath = self.parent.manager.PROJECTS
        self.projName = tk.StringVar()
        self.projName.set('')
        self.projLocation = tk.StringVar()
        self.projLocation.set(self.projectsPath)

        content = tk.Frame(self.root)
        content.columnconfigure(0, weight=1)
        content.columnconfigure(1, weight=3)
        content.config(bg='white')
        content.grid(row=0, column=0, sticky='news', padx=5, pady=5)

        #  Project name line
        labelName = tk.Label(content, text=Message.LABEL_PROJECT + ' name', bg='white', bd=0)
        labelName.grid(row=0, sticky=tk.W, padx=5, pady=5)
        entryName = tk.Entry(content, bg=cfgEntryBgColor, width=20, textvariable=self.projName)
        entryName.grid(row=0, column=1, columnspan=2, sticky=tk.W, padx=5, pady=5)
        entryName.bind("<Return>", self._create)
        entryName.bind("<KP_Enter>", self._create)
        # Project location line
        labelLocation = tk.Label(content, text=Message.LABEL_PROJECT + ' location', bg='white', bd=0)
        labelLocation.grid(row=1, column=0, sticky='nw', padx=5, pady=5)

        self.entryBrowse = tk.Entry(content, bg=cfgEntryBgColor, width=40, textvariable=self.projLocation)
        self.entryBrowse.grid(row=1, column=1, sticky='nw', padx=5, pady=5)
        self.btnBrowse = IconButton(content, 'Browse', Icon.ACTION_BROWSE,
                                    highlightthickness=0, command=self._browsePath)
        self.btnBrowse.grid(row=1, column=2, sticky='e', padx=5, pady=5)

        self.initial_focus = entryName
        self.initial_focus.focus()

        btnFrame = tk.Frame(content)
        btnFrame.columnconfigure(0, weight=1)
        btnFrame.grid(row=2, column=0, sticky='sew', padx=5, pady=(0, 5), columnspan=2)
        btnFrame.config(bg='white')

        # Create buttons
        btnCreate = HotButton(btnFrame, 'Create', Icon.BUTTON_SELECT, command=self._create)
        btnCreate.grid(row=0, column=0, sticky='e', padx=5, pady=5)
        btnCancel = Button(btnFrame, 'Cancel', Icon.BUTTON_CANCEL, command=self.close)
        btnCancel.grid(row=0, column=1, sticky='e', padx=5, pady=5)

    def _browsePath(self, e=None):
        def onSelect(obj):
            self.projLocation.set(obj.getPath())

        v = self.projLocation.get().strip()
        path = None
        if v:
            v = os.path.dirname(v)
            if os.path.exists(v):
                path = v
        if not path:
            path = self.projectsPath

        browser = FileBrowserWindow("Browsing", self, path=path, onSelect=onSelect, onlyFolders=True)
        browser.show()

    def _create(self, e=None):
        projName = self.projName.get().strip()
        projLocation = self.projLocation.get().strip()

        # Validate that project name is not empty
        if not projName:
            showError("Validation error", "Project name is empty", self.root)
        # Validate that project location is not empty
        elif not projLocation:
            showError("Validation error", "Project location is empty", self.root)
        # Validate that project location exists
        elif not os.path.exists(projLocation):
            showError("Validation error", "Project location does not exist", self.root)
        # Validate that project location is a directory
        elif not os.path.isdir(projLocation):
            showError("Validation error", "Project location is not a directory", self.root)
        # Validate that project path (location + name) ddoes not exists
        elif os.path.exists(os.path.join(projLocation, projName)):
            showError("Validation error", "Project path already exists", self.root)
        else:
            self.parent.createNewProject(projName, projLocation)
            self.close()


class ProjectImportWindow(Window):
    """ Windows to import a project. """
    def __init__(self, title, parent=None, weight=True, minsize=(400, 150),
                 icon=Icon.SCIPION_ICON, **args):
        """
         We assume the parent should be ProjectsView
        """
        Window.__init__(self, title, parent.windows, weight=weight,
                        icon=icon, minsize=minsize, enableQueue=True)
        self.root['background'] = 'white'
        self.parent = parent
        # Dirty hack, need to add a slash for the explorer to pick up the right default path.
        self.projectsPath = getHomePath()+"/"
        self.projLocation = tk.StringVar()
        self.projLocation.set(self.projectsPath)

        self.projName = tk.StringVar()
        self.projName.set('')

        self.searchLocation = tk.StringVar()
        self.searchLocation.set('')

        content = tk.Frame(self.root)
        content.columnconfigure(0, weight=1)
        content.columnconfigure(1, weight=1)
        content.config(bg='white')
        content.grid(row=0, column=0, sticky='news',
                     padx=5, pady=5)

        # Path explorer
        labelProjectLocation = tk.Label(content, text="Project location", bg='white', bd=0)
        labelProjectLocation.grid(row=0, column=0, sticky='nw', padx=5, pady=5)
        # it seems tk.Entry does not uses default font...grrrr!!
        self.entryBrowse = tk.Entry(content, bg=cfgEntryBgColor, width=40,
                                    textvariable=self.projLocation, font=self.font)
        self.entryBrowse.grid(row=0, column=1, sticky='nw', padx=5, pady=5)
        self.btnBrowse = IconButton(content, 'Browse', Icon.ACTION_BROWSE, highlightthickness=0,
                                    command=self._browseProjectLocation)
        self.btnBrowse.grid(row=0, column=2, sticky='e', padx=5, pady=5)

        # Copy files check
        labelCheck = tk.Label(content, text="Copy project", bg='white', borderwidth=0)
        labelCheck.grid(row=1, column=0, sticky='nw', padx=5, pady=5)

        self.tkCheckVar = tk.IntVar()
        btnCheck = tk.Checkbutton(content, variable=self.tkCheckVar, highlightthickness=0, activebackground='white',
                                  bg='white', bd=0)
        btnCheck.grid(row=1, column=1, sticky='nw', padx=0, pady=5)

        btnCopyHelp = IconButton(content, Message.LABEL_BUTTON_HELP, Icon.ACTION_HELP, highlightthickness=0,
                                 command=lambda: self.showInfo('If checked, \"Project location\" will be copied. Otherwise a soft link to it will be created.'))
        btnCopyHelp.grid(row=1, column=3, sticky='e', padx=2, pady=2)

        # Project name
        labelName = tk.Label(content, text='Project name (Optional)', bg='white', bd=0)
        labelName.grid(row=2, column=0, sticky='nw', padx=5, pady=5)
        entryName = tk.Entry(content, bg='white', width=20, textvariable=self.projName, font=self.font)
        entryName.grid(row=2, column=1, sticky='nw', padx=5, pady=5)

        # Path to search for raw data and restore broken links.
        labelSearchLocation = tk.Label(content, text="Raw files location (Optional)", bg='white', bd=0)
        labelSearchLocation.grid(row=3, column=0, sticky='nw', padx=5, pady=5)

        self.entrySearchLocation = tk.Entry(content, bg='white', width=40,
                                            textvariable=self.searchLocation, font=self.font)
        self.entrySearchLocation.grid(row=3, column=1, sticky='nw', padx=5, pady=5)
        self.btnSearch = IconButton(content, 'Browse', Icon.ACTION_BROWSE,
                                    highlightthickness=0, command=self._browseSearchLocation)
        self.btnSearch.grid(row=3, column=2, sticky='e', padx=5, pady=5)
        btnSearchHelp = IconButton(content, Message.LABEL_BUTTON_HELP, Icon.ACTION_HELP, highlightthickness=0,
                                   command=lambda: self.showInfo('Optional: Folder where raw files, binaries (movies, micrographs,..) can be found. Used to repair broken links.'))
        btnSearchHelp.grid(row=3, column=3, sticky='e', padx=2, pady=2)

        self.initial_focus = entryName
        self.initial_focus.focus()
        btnCheck.select()

        btnFrame = tk.Frame(content)
        btnFrame.columnconfigure(0, weight=1)
        btnFrame.grid(row=4, column=0, sticky='sew', padx=5, pady=(0, 5), columnspan=2)
        btnFrame.config(bg='white')

        # Create buttons
        btnSelect = HotButton(btnFrame, 'Import', Icon.BUTTON_SELECT, command=self._select)
        btnSelect.grid(row=0, column=0, sticky='e', padx=5, pady=5)
        btnCancel = Button(btnFrame, 'Cancel', Icon.BUTTON_CANCEL, command=self.close)
        btnCancel.grid(row=0, column=1, sticky='e', padx=5, pady=5)

    def _browseProjectLocation(self, e=None):
        self._browsePath(self.projLocation)

    def _browseSearchLocation(self, e=None):
        self._browsePath(self.searchLocation)

    def _browsePath(self, location):
        def onSelect(obj):
            location.set(obj.getPath())

        v = location.get().strip()
        path = None
        if v:
            v = os.path.dirname(v)
            if os.path.exists(v):
                path = v
        if not path:
            path = self.projectsPath

        browser = FileBrowserWindow("Browsing",
                                    self, path=path,
                                    onSelect=onSelect,
                                    onlyFolders=True)
        browser.show()

    def _select(self):
        projName = self.projName.get().strip()
        projLocation = self.projLocation.get().strip()
        copyFiles = self.tkCheckVar.get() != 0
        searchLocation = self.searchLocation.get().strip()
        manager = Manager()

        # If project name is empty we will use the same name as the source
        if not projName:
            projName = os.path.basename(projLocation)

        errorMessage = ''

        # Validate that project location is not empty
        if not projLocation:
            errorMessage = "Project location is empty\n"

        # Validate that project location exists
        elif not os.path.exists(projLocation):
            errorMessage += "Project location does not exist\n"

        # Validate that project location is a directory
        elif not os.path.isdir(projLocation):
            errorMessage += "Project location is not a directory\n"
        # Validate that the project location is a scipion project folder
        elif not os.path.exists(os.path.join(projLocation, Project.getDbName())):
            errorMessage += "Project location doesn't look like a scipion folder\n"

        # Validate that there isn't already a project with the same name
        if manager.hasProject(projName):
            errorMessage += "Project [%s] already exists\n" % projName

        # Validate that search location exists
        if searchLocation:
            if not os.path.exists(searchLocation):
                errorMessage += "Raw files location does not exist\n"
            # Validate that search location is a directory
            elif not os.path.isdir(searchLocation):
                errorMessage += "Raw files location is not a directory\n"

        if errorMessage:
            showError("Validation error", errorMessage, self.root)
        else:
            self.parent.importProject(projLocation, copyFiles, projName, searchLocation)
            self.close()



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
# JMRT (2018-12-11) This module is almost not used at all. Maybe it
can be removed in a future. Just kept here for the moment inside
pw.utils and not imported by default
"""

import os
from collections import OrderedDict, namedtuple

from pyworkflow.mapper import SqliteFlatDb, SqliteDb


class DataSet(object):
    """ Holds several Tables
    All tables should have an unique tableName. 
    """
    
    def __init__(self, tables, tableName=None, volumeName=None, numberSlices=0):
        self._tables = list(tables)
        self._tableName = tableName
        # NAPA de LUXE: Hay que ver si el volumen name se usa en algun lado
        self._volumeName = volumeName 
        self._numberSlices = numberSlices 
        self.projectPath = None
        
    def currentTable(self):
        """ Returns the name of the last selected table. """    
        return self._tableName
        
    def setVolumeName(self, volumeName):    
        self._volumeName = volumeName
        
    def getVolumeName(self):
        return self._volumeName
    
    def setNumberSlices(self, numberSlices):
        self._numberSlices = numberSlices
        
    def getNumberSlices(self):
        return self._numberSlices    
    
    def getNumberSlicesForTemplate(self):
        return range(self._numberSlices)
        
    def listTables(self):
        """ List the actual table names on the DataSet. """
        return self._tables
    
    def getTable(self, tableName=None):
        if tableName is None:
            tableName = self.listTables()[0]
        if tableName not in self._tables:
            raise Exception("DataSet: table '%s' not found.\n   Current tables: %s" % 
                            (tableName, self._tables))
            
        table = self._loadTable(tableName)
        self._tableName = tableName
        
        return table
    
    def getTypeOfColumn(self, label):
        """ this method should be implemented by subclasses. """
        pass
    
    def _loadTable(self, tableName):
        """ this method should be implemented by subclasses. """
        pass


class Table(object):
    """ Table to hold rows of data. 
    A table contains a list of columns. """
    def __init__(self, *columns):
        self._columns = OrderedDict()
        self._rowDict = OrderedDict()
        self._addColumn(Column('id', int))
        
        for col in columns:
            self._addColumn(col)
            
        colNames = [col.getName() for col in self.iterColumns()]
        # This imply that columns can only be added in __init__
        self.Row = namedtuple('Row', colNames)
    
    def setLabelToRender(self, labelToRender):
        self._labelToRender = labelToRender
        
    def _addColumn(self, col):
        self._columns[col.getName()] = col
        
    def getColumnValues(self, columnName):
        if self.hasColumn(columnName):
            return [getattr(row, columnName) for row in self.iterRows()] 
        else:
            return [None] * self.getSize()
        
    def iterColumns(self):
        return self._columns.values()
    
    def hasColumn(self, columnName):
        """ Return true if column exists """
        return columnName in self._columns
    
    def getColumn(self, columnName):
        if columnName not in self._columns:
            raise Exception('Table: column "%s" not found.\Current columns: %s' % (
                columnName, '\n'.join(self._columns.keys())))
        return self._columns[columnName] 
    
    def hasEnabledColumn(self):
        """ Return true if enabled column exists """
        return self.hasColumn('enabled')
        
    def getColumns(self):
        """ Return all columns. """
        return self._columns.values()
    
    def getNumberOfColumns(self):
        return len(self._columns)
    
    def getSize(self):
        """ Return the number of rows. """
        return len(self._rowDict)
    
    def getRows(self):
        """ Return all rows. """
        return [row for row in self.iterRows()]
    
    def getRow(self, rowId):
        return self._rowDict[rowId]
    
    def _setRow(self, rowId, row):
        self._rowDict[rowId] = row
    
    def getDataToRenderAndExtra(self):
        return zip(self.getIdColumn(),
                   self.getColumnValues("enabled"),
                   self.getDataToRender(),
                   self.getTransformationMatrix())
    
    def getDataToRender(self):
        return self.getColumnValues(self._labelToRender)
    
    def getIdColumn(self):
        return self.getColumnValues("id")
        
    def getTransformationMatrix(self):    
        return self.getColumnValues(self._labelToRender+"_transformationMatrix")
    
    def _convertValues(self, values):
        """ Convert the input values to the actual
        expected type of each column. 
        """
        cValues = {}
        for k, v in values.items():
            col = self.getColumn(k)
            cValues[k] = col.convert(v)
        
        return cValues
        
    def addRow(self, rowId, **values):
        """ With this implementation the rowId should be provided.
        We need to work around to also allow automatic generation of id's
        """
        values['id'] = rowId
        
        for col in self.iterColumns():
            if col.getName() not in values:
                if col.hasDefault():
                    values[col.getName()] = col.getDefault()
                else:
                    raise Exception('Table: value for column "%s" not provided.' % col.getName())
                
        row = self.Row(**self._convertValues(values))
        self._setRow(rowId, row)
        
    def updateRow(self, rowId, **values):
        """ Update a row given its rowId and some values to update. """
        row = self.getRow(rowId)
        self._setRow(rowId, row._replace(**self._convertValues(values)))
    
    def iterRows(self):
        """ Iterate over the rows. """
        return self._rowDict.values()
    
    def getValueFromIndex(self, index, label):
        """ Return the value of the property 'label'
        in the element that has this 'index'.
        """
        value = list(self._rowDict.values())[index]._asdict()[label]
        return value
    
    def getIndexFromValue(self, value, label):
        """ Search the element that has property 'label'
        equals to value and returns its index.
        """
        for index, row in enumerate(self.iterRows()):
            if value == row._asdict()[label]:
                return index            
        return -1
    
    def __str__(self):
        return '\n'.join([str(row) for row in self.iterRows()])
    

# JMRT (2018-12-11) This constants are duplicated in showj, since this module
# is not widely used I don't find convenient such a dependency
COL_RENDER_NONE = 0
COL_RENDER_ID = 1
COL_RENDER_TEXT = 2
COL_RENDER_IMAGE = 3
COL_RENDER_CHECKBOX = 4
COL_RENDER_VOLUME = 5


class Column(object):
    def __init__(self, colName, colType=None, default=None, 
                 label=None, renderType=COL_RENDER_NONE):
        self._name = colName
        self._type = colType
        self._default = default
        self._label = label or colName
        self._renderType = renderType
        
    def getName(self):
        return self._name

    def getLabel(self):
        return self._label
        
    def getType(self):
        return self._type
    
    def convert(self, value):
        """ Try to convert the value to the column type. """
        return self._type(value)
    
    def hasDefault(self):
        return self._default is not None
    
    def getDefault(self):
        return self._default
    
    def getRenderType(self):
        return self._renderType
    
    def setRenderType(self, renderType):
        self._renderType = renderType
    

class SqliteDataSet(DataSet):
    """ Provide a DataSet implementation based on sqlite file.
    The tables of the dataset will be the object tables in database.
    Each block is a table on the dataset. 
    """
    
    def __init__(self, filename):
        self._dbName = filename
        db = SqliteDb()
        db._createConnection(filename, 1000)
        # Tables should be at pairs:
        # PREFIX_Classes
        # PREFIX_Objects  
        # where PREFIX can be empty
        self.tablePrefixes = OrderedDict()
        tables = db.getTables()
        for t in tables:
            if t.endswith('Classes'):
                prefix = t.replace('Classes', '')
                to = prefix + 'Objects'
                if to not in tables:
                    raise Exception('SqliteDataSet: table "%s" found, but not "%s"' % (t, to))
                flatDb = SqliteFlatDb(filename, tablePrefix=prefix)
                tableName = prefix + self._getPlural(flatDb.getSelfClassName())
                self.tablePrefixes[tableName] = prefix
                # tablePrefixes.append(prefix)
        DataSet.__init__(self, self.tablePrefixes.keys())
        db.close()
        
    def _getPlural(self, className):
        """ Get the plural of word for tables labels. """
        if className.startswith('Class'):
            return className.replace('Class', 'Classes')
        return className + 's'
        
    def _loadTable(self, tableName):
        """ Load information from tables PREFIX_Classes, PREFIX_Objects. """
        
        tableName = self.tablePrefixes[tableName]
        
        BASIC_COLUMNS = [Column('id', int, renderType=COL_RENDER_ID), 
                         Column('enabled', bool, renderType=COL_RENDER_CHECKBOX),
                         Column('label', str), 
                         Column('comment', str),
                         Column('creation', str)]
        # Load columns from PREFIX_Classes table
        columns = list(BASIC_COLUMNS)
        db = SqliteDb()
        db._createConnection(self._dbName, 1000)
        db.executeCommand("SELECT * FROM %sClasses;" % tableName)
        # This will store the images columsn to join
        # the _index and the _filename
        imgCols = {}
        for row in db._iterResults():
            renderType = COL_RENDER_NONE
            colName = row['column_name']
            colLabel = row['label_property']
            
            if colLabel != 'self':
                # Keep track of _index and _filename pairs to mark as renderable images
                if colLabel.endswith('_index'):
                    imgCols[colLabel.replace('_index', '')] = colName
                
                elif colLabel.endswith('_filename'):
                    
                    # TODO: Maybe not all the labels endswith "_filename" 
                    # have to be rendered. 
                    # for example in the RotSpectra with '_representative._filename'

                    prefix = colLabel.replace('_filename', '')
                    if prefix in imgCols:
                        renderType = COL_RENDER_IMAGE
                        imgCols[colName] = imgCols[prefix]
                
                # CTF FIX
                elif (colLabel.endswith('_psdFile') or 
                      colLabel.endswith('_enhanced_psd') or 
                      colLabel.endswith('_ctfmodel_quadrant') or 
                      colLabel.endswith('_ctfmodel_halfplane')):
                    
                    renderType = COL_RENDER_IMAGE
                
                if row['class_name'] == 'Boolean':
                    renderType = COL_RENDER_CHECKBOX   
                columns.append(Column(colName, str, label=colLabel, renderType=renderType))
        table = Table(*columns)
        
        checkedImgCols = {}  # Check if the image columns are volumes
        # FIXME: Move this to scipion-em? Maybe remove the whole module that is not used?
        from pwem.emlib.image import ImageHandler
        ih = ImageHandler() 
        
        # Populate the table in the DataSet
        db.executeCommand("SELECT * FROM %sObjects;" % tableName)
        for row in db._iterResults():
            rowDict = dict(row)
            for k, v in rowDict.items():
                if v is None:
                    rowDict[k] = ''
                # Set the index@filename for images columns values
                if k in imgCols:
                    colName = imgCols[k]
                    index = rowDict[colName]
                    
                    filename = os.path.join(self.projectPath, rowDict[k])
                    filepath = filename.replace(":mrc", "")
                    if not checkedImgCols.get(colName, False):
                        if os.path.exists(filepath):
                            # print "Fn to get dims: %s@%s" % (index,filename)
                            x, y, z, n = ih.getDimensions((index, filename))
                            if z > 1:
                                table.getColumn(k).setRenderType(COL_RENDER_VOLUME)
                        checkedImgCols[colName] = True
                    if index:
                        rowDict[k] = '%06d@%s' % (index, filename)
            table.addRow(row['id'], **rowDict)
            
        return table
        
        
class SingleFileDataSet(DataSet):
    """ DataSet implementation for single files such as Images or Volumes. 
    """
    
    def __init__(self, filename):
        self._filename = filename
        self._tableName = ""
        DataSet.__init__(self, [self._tableName])
        self._table = self._createSingleTable()
        
    def _createSingleTable(self):
        table = Table(Column('filename', str, 
                             renderType=COL_RENDER_VOLUME))  # FIXME: for single images we need to read the dimensions
        table.addRow(1, filename=self._filename)
        
        return table
        
    def _loadTable(self, tableName):
        return self._table
